from __future__ import annotations
from collections import OrderedDict
from pathlib import Path
import textwrap
from typing import Union
import importlib
import black
from psychonaut.lexicon.constants import MODULE_PREFIX
from psychonaut.lexicon.ctx import GenCtx
from psychonaut.lexicon.fields import generate_pydantic_fields
from psychonaut.lexicon.planner import LexiconDef, compile_toposorted_lexicon_defns
from psychonaut.lexicon.types import (
    LexObject,
    LexRecord,
    LexRef,
    LexRefUnion,
    LexString,
    LexToken,
    LexXrpcParameters,
    LexXrpcProcedure,
    LexXrpcQuery,
)
from .util import (
    camel_to_class_name,
    camel_to_snake,
    import_types_str,
    build_python_module_file_structure,
)
from black import format_str, FileMode


def generate_all(
    input_dir: Path,
    output_dir: Path,
    verbose: bool = False,
    import_all_test: bool = False,
):
    # Create the output directory if it doesn't exist.
    # The ensure_path_and_init_files function will do this for us, but
    # it's nice to visibly mark progress.
    output_dir.mkdir(parents=True, exist_ok=True)

    # With the way we're generating and collecting code fragments,
    # we don't actually need to do a topological sort. But
    # doing so does some really helpful sanity checks early on.
    ctxs = OrderedDict()
    global_vars = {}
    global_vars['no_import'] = set()
    for node in compile_toposorted_lexicon_defns(input_dir):
        if verbose:
            # print(f"Processing {node.node_id} in {node.input_path}")
            pass

        # Generation contexts are at a per-file granularity.
        ctx = ctxs.get(node.input_path)
        if not ctx:
            ctx = GenCtx(
                verbose=verbose, input_path=node.input_path, global_vars=global_vars
            )
            ctx.imports["typing"].add("Any")  # TODO: ugly kludge
            ctxs[node.input_path] = ctx

        _generate_defn(ctx, node)

    # Write out all the generated code.
    # There are a few points during which things could still fail.
    # But having partial output is kinda convenient for debugging.
    module_paths = []
    for input_path, ctx in ctxs.items():
        module_paths.append(_write_generated(ctx, output_dir, verbose=verbose))

    if not import_all_test:
        return

    # It's a bit easier to try over all of them and then fail with the codegen.
    did_fail = False
    for module_path in module_paths:
        try:
            importlib.import_module(module_path)
        except Exception as e:
            print(f"!Failed to import {module_path}: {e}")
            did_fail = True
    
    if did_fail:
        raise Exception("Failed to import all generated modules")



# = Level 1 ==================================================================


def _generate_defn(ctx: GenCtx, node: LexiconDef):
    # The following is roughly in order of use frequency.
    # python (3.10) match statements aren't implemented with jump tables
    # so this prob has no effect on performance. But it does hint
    # at the order of importance for covering a lot of the lexicon.
    match node.defn.type:
        case "query":
            generate_query(ctx, node)
        case "procedure":
            generate_procedure(ctx, node)
        case "record":
            generate_record(ctx, node)
        case "object":
            generate_object(ctx, node)
        case "subscription":
            pass
        case "token":
            generate_token(ctx, node)
        case "string":
            generate_string(ctx, node)
        case _:
            raise Exception(f"Unknown defn type: {node.defn.type} in {node}")


def _write_generated(ctx: GenCtx, output_dir: Path, verbose: bool) -> str:
    code_blocks = [generate_imports(ctx)] + [
        frag.generated_src for frag in ctx.fragments.values()
    ]

    src = "\n".join(code_blocks)
    if not src.strip() or not ctx.common_xrpc_id:  # TODO: should never happen
        return

    output_path = build_python_module_file_structure(output_dir, ctx.common_xrpc_id)

    if verbose:
        print(f"\tWriting generated code to {output_path} {ctx.common_xrpc_id}")

    with open(output_path, "w") as f:
        # Black formatted. This is good because,
        # - it enforces a consistent style including in regenerated commits
        # - block does syntax validation so this is another sanity check
        black_formatted = ""
        try:
            black_formatted = format_str(src, mode=FileMode())
        except black.InvalidInput:
            print(f"Failed to format {output_path}")
            print(src)
            raise

        f.write(black_formatted)

    return MODULE_PREFIX + camel_to_snake(ctx.common_xrpc_id)


# = Level 2 ==================================================================


def _setup_and_validate(
    ctx: GenCtx,
    node: LexiconDef,
    expected_defn_type: type,
    log_message: str,
    skip_deps: bool = False,
):
    assert isinstance(
        node.defn, expected_defn_type
    ), f"{expected_defn_type.__name__} != {type(node.defn)}"

    if ctx.verbose:
        print(f"{log_message} {node.node_id}: {node.leaf_id_as_class_name}")

    if not skip_deps:
        _add_dep_imports(ctx, node)


def generate_query(ctx: GenCtx, node: LexiconDef):
    _setup_and_validate(ctx, node, LexXrpcQuery, "Generating query")

    # Queries use a Session object
    ctx.imports["psychonaut.api.session"].add("Session")

    blocks = []

    # If there is an output, generate the Resp first
    resp_name = "Any"
    if output := node.defn.output:
        if isinstance(output.schema_kludge, LexObject):
            resp_name = node.leaf_id_as_class_name + "Resp"
            desc = output.description or output.schema_kludge.description
            blocks.append(
                generate_pydantic_model(
                    ctx,
                    resp_name,
                    desc,
                    output.schema_kludge or {},
                )
            )
        elif isinstance(output.schema_kludge, LexRef):
            print("REF", output)
        elif isinstance(output.schema_kludge, LexRefUnion):
            print("REF UNION", output)
        else:
            print("UNKNOWN", output)

    # There must always be a request class for Psychonaut
    req_name = node.leaf_id_as_class_name + "Req"
    blocks.append(
        generate_pydantic_model(
            ctx,
            req_name,
            node.defn.description or "[none provided by spec]",
            node.defn.parameters or {},
        )
    )
    blocks.append(_generate_xrpc_id_property(node))

    blocks.append(
        build_async_session_f(
            ctx,
            "query",
            resp_name,
        )
    )

    node.generated_src = "\n\n".join(blocks)

    ctx.add_fragment(req_name, node)


def generate_procedure(ctx: GenCtx, node: LexiconDef):
    _setup_and_validate(ctx, node, LexXrpcProcedure, "Generating procedure")

    # Procedures use a Session object
    ctx.imports["psychonaut.api.session"].add("Session")

    # Procedures area little trickier than quries because
    # they may have both parameters and an input.
    assert not (
        node.defn.input and node.defn.parameters
    ), f"procedure {node.node_id} input and parameters"
    req_schema = {}
    if node.defn.input:
        req_schema = node.defn.input.schema_kludge
    elif node.defn.parameters:
        req_schema = node.defn.parameters.schema_json

    blocks = []

    # If there is an output, generate the Resp first
    resp_name = "Any"
    if output := node.defn.output:
        if isinstance(output.schema_kludge, LexObject):
            resp_name = node.leaf_id_as_class_name + "Resp"
            desc = output.description or output.schema_kludge.description
            blocks.append(
                generate_pydantic_model(
                    ctx,
                    resp_name,
                    desc,
                    output.schema_kludge or {},
                )
            )
        elif isinstance(output.schema_kludge, LexRef):
            print("REF", output)
        elif isinstance(output.schema_kludge, LexRefUnion):
            print("REF UNION", output)
        else:
            print("UNKNOWN", output)

    # There must always be a request class for Psychonaut
    req_name = node.leaf_id_as_class_name + "Req"

    blocks.append(
        generate_pydantic_model(
            ctx,
            req_name,
            node.defn.description or "[none provided by spec]",
            req_schema,
        )
    )
    blocks.append(_generate_xrpc_id_property(node))
    blocks.append(
        build_async_session_f(
            ctx,
            "procedure",
            resp_name,
        )
    )

    node.generated_src = "\n\n".join(blocks)
    ctx.add_fragment(req_name, node)


def generate_record(ctx: GenCtx, node: LexiconDef):
    _setup_and_validate(ctx, node, LexRecord, "Generating record")

    blocks = []
    class_name = node.leaf_id_as_class_name
    blocks.append(
        generate_pydantic_model(
            ctx,
            node.leaf_id_as_class_name,
            node.defn.description or "[none provided by spec]",
            node.defn.record or {},
        )
    )
    blocks.append(_generate_xrpc_id_property(node))

    node.generated_src = "\n\n".join(blocks)
    ctx.add_fragment(class_name, node)


def generate_object(ctx: GenCtx, node: LexiconDef):
    _setup_and_validate(ctx, node, LexObject, "Generating object")

    # app.bsky.embed.externa.json is annoying because the main is an object and so is the external
    # TODO: figure out how to handle this
    obj_name = node.leaf_id_as_class_name
    if node.node_id == "app.bsky.embed.external" and len(node.defn.properties) == 1:
        obj_name = "ExternalFML"

    blocks = []
    blocks.append(
        generate_pydantic_model(
            ctx,
            node.leaf_id_as_class_name,
            node.defn.description or "[none provided by spec]",
            node.defn or {},
        )
    )

    node.generated_src = "\n\n".join(blocks)
    ctx.add_fragment(obj_name, node)


def generate_string(ctx: GenCtx, node: LexiconDef):
    _setup_and_validate(ctx, node, LexString, "Generating object", skip_deps=True)
    ctx.imports["enum"].update(["Enum", "auto"])

    lines = [f"class {node.leaf_id_as_class_name}(str, Enum):"]
    for dep in node.deps:
        # TODO: add description
        tok = ctx.global_vars[dep]
        as_constant = camel_to_class_name(tok.leaf_id).upper()
        lines.append(f"    {as_constant} = {repr(tok.leaf_id)}")

    node.generated_src = "\n".join(lines)
    ctx.add_fragment(node.leaf_id_as_class_name, node)


def generate_token(ctx: GenCtx, node: LexiconDef):
    assert isinstance(node.defn, LexToken), f"LexToken != {type(node.defn)}"
    if ctx.verbose:
        print(f"Generating token {node.node_id}: {node.leaf_id_as_class_name}")

    ctx.global_vars[node.node_id] = node
    ctx.global_vars['no_import'].add(node.node_id)


# = Level 3 ==================================================================


def _generate_xrpc_id_property(node: LexiconDef) -> str:
    return f"""
    @property
    def xrpc_id(self) -> str:
        return "{node.node_id}"
    """


def generate_pydantic_model(
    ctx: GenCtx,
    class_name: str,
    docstr: str,
    params: Union[LexXrpcParameters, LexObject],
    parent_class: str = "BaseModel",
) -> str:
    """
    Generate a pydantic model class from the given parameters.

    :param ctx: The generation context (save things like imports)
    :param class_name: The name of the class to generate
    :param docstr: The docstring for the class
    :param params: The parameters to generate the class from
    """
    model_lines = []

    # Signal that we need to import pydantic's basemodel
    ctx.imports["pydantic"].add(parent_class)

    # Generate the pydantic fields (if any)
    field_lines = generate_pydantic_fields(ctx, params)
    if field_lines:
        ctx.imports["pydantic"].add("Field")

    model_lines.append(f"class {class_name}(BaseModel):")
    if docstr:
        model_lines.append(generate_class_docstr(docstr))

    for field_line in field_lines:
        model_lines.append(f"    {field_line}")

    model_lines.append("")

    return "\n".join(model_lines)


def generate_class_docstr(docstr: str) -> str:
    return "\n".join(
        [
            '    """',
            textwrap.fill(
                docstr, width=76, initial_indent="    ", subsequent_indent="    "
            ),
            '    """',
        ]
    )


def generate_imports(ctx: GenCtx) -> str:
    model_lines = []
    names_seen, shadow_detected = set(), False
    for import_path, type_names in ctx.imports.items():
        model_lines.append(import_types_str(import_path, type_names))
        for name in type_names:
            if name in names_seen:
                shadow_detected = True
            names_seen.add(name)

    if shadow_detected:
        print(f"WARNING: shadowing imports detected in {ctx.common_xrpc_id}")
    model_lines.append("\n")

    return "\n".join(model_lines)


def build_async_session_f(ctx: GenCtx, f: "query", resp_class: str) -> str:
    ctx.imports["psychonaut.api.session"].add("Session")
    model_lines = []
    model_lines.append(f"    async def do_xrpc(self, sess: Session) -> {resp_class}:")

    if resp_class == "Any":
        model_lines.append(f"        return await sess.{f}(self)")
    else:
        model_lines.append(f"        resp = await sess.{f}(self)")
        model_lines.append(f"        return {resp_class}(**resp)")
    return "\n".join(model_lines)


def _add_dep_imports(ctx: GenCtx, node: LexiconDef):
    for dep in node.deps:
        if "#" not in dep:
            continue
        module, obj = dep.split("#")

        if module == ctx.common_xrpc_id:
            continue

        if dep in ctx.global_vars['no_import']:
            continue

        module = camel_to_snake(module)
        ctx.imports[MODULE_PREFIX + module].add(camel_to_class_name(obj))



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_identifier


class GetProfileReq(BaseModel):
    """
    [none provided by spec]
    """

    actor: str = Field(..., pre=True, validator=validate_at_identifier)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.actor.getProfile"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did


class CreateInviteCodeResp(BaseModel):
    code: str = Field(...)


class CreateInviteCodeReq(BaseModel):
    """
    Create an invite code.
    """

    useCount: int = Field(...)
    forAccount: Optional[str] = Field(default=None, pre=True, validator=validate_did)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.createInviteCode"

    async def do_xrpc(self, sess: Session) -> CreateInviteCodeResp:
        resp = await sess.procedure(self)
        return CreateInviteCodeResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetModerationActionReq(BaseModel):
    """
    View details about a moderation action.
    """

    id: int = Field(...)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.getModerationAction"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any
from enum import auto, Enum


class ReasonType(str, Enum):
    REASONVIOLATION = "reasonViolation"
    REASONOTHER = "reasonOther"
    REASONSPAM = "reasonSpam"
    REASONMISLEADING = "reasonMisleading"
    REASONRUDE = "reasonRude"
    REASONSEXUAL = "reasonSexual"



from typing import Any, Optional, List
from psychonaut.api.lexicons.app.bsky.feed.defs import PostView
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_array


class GetPostsResp(BaseModel):
    posts: Any


class GetPostsReq(BaseModel):
    """
    A view of an actor's feed.
    """

    uris: List[str] = Field(
        ..., max_items=25, pre=True, validator=validate_array(validate_at_uri)
    )

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.getPosts"

    async def do_xrpc(self, sess: Session) -> GetPostsResp:
        resp = await sess.query(self)
        return GetPostsResp(**resp)



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileView
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid


class GetRepostedByResp(BaseModel):
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)
    cursor: Optional[str] = Field(default=None)
    repostedBy: Any


class GetRepostedByReq(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)
    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.getRepostedBy"

    async def do_xrpc(self, sess: Session) -> GetRepostedByResp:
        resp = await sess.query(self)
        return GetRepostedByResp(**resp)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_cid, validate_did


class GetCheckoutReq(BaseModel):
    """
    Gets the repo state.
    """

    did: str = Field(
        ..., description="The DID of the repo.", pre=True, validator=validate_did
    )
    commit: Optional[str] = Field(
        default=None,
        description="The commit to get the checkout from. Defaults to current HEAD.",
        pre=True,
        validator=validate_cid,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.getCheckout"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any, Optional, List
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileViewDetailed
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_identifier, validate_array


class GetProfilesResp(BaseModel):
    profiles: Any


class GetProfilesReq(BaseModel):
    """
    [none provided by spec]
    """

    actors: List[str] = Field(
        ..., max_items=25, pre=True, validator=validate_array(validate_at_identifier)
    )

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.actor.getProfiles"

    async def do_xrpc(self, sess: Session) -> GetProfilesResp:
        resp = await sess.query(self)
        return GetProfilesResp(**resp)



import asyncio
from typing import Optional
import aiodns


SUBDOMAIN = "_atproto"
PREFIX = "did="


class NoHandleRecordError(Exception):
    pass


async def resolve_dns(
    handle: str, resolver: Optional[aiodns.DNSResolver] = None
) -> str:
    resolver = resolver or aiodns.DNSResolver()

    try:
        resp = await resolver.query(f"{SUBDOMAIN}.{handle}", "TXT")

        first_did = next((r.text for r in resp if r.text.startswith(PREFIX)), None)
        if first_did:
            return first_did[len(PREFIX) :]
    except aiodns.error.DNSError as err:
        if err.args[0] != aiodns.error.ARES_ENOTFOUND:
            raise err
    raise NoHandleRecordError()



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_handle, validate_did


class ResolveHandleResp(BaseModel):
    did: str = Field(..., pre=True, validator=validate_did)


class ResolveHandleReq(BaseModel):
    """
    Provides the DID of a repo.
    """

    handle: Optional[str] = Field(
        default=None,
        description="The handle to resolve. If not supplied, will resolve the host's own handle.",
        pre=True,
        validator=validate_handle,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.identity.resolveHandle"

    async def do_xrpc(self, sess: Session) -> ResolveHandleResp:
        resp = await sess.query(self)
        return ResolveHandleResp(**resp)



"""
https://github.com/bluesky-social/atproto/blob/d8b50c73e4915277da049e18cd5f6296d978aff4/packages/common-web/src/ipld.ts
"""
from multiformats.cid import CID
from typing import Any, Dict, List, Optional, Union
import base64

# TODO: untested mostly

JsonValue = Union[bool, float, str, None, Any, List[Any], Dict[str, Any]]
IpldValue = Union[JsonValue, CID, bytes, List[Any], Dict[str, Any]]


def as_cid(obj) -> Optional[CID]:
    if isinstance(obj, CID):
        return obj

    if isinstance(obj, (str, bytes)):
        try:
            return CID.decode(obj)
        except (ValueError, KeyError):
            return None

    return None


def json_to_ipld(val: JsonValue) -> IpldValue:
    if isinstance(val, list):
        return [json_to_ipld(item) for item in val]

    if val and isinstance(val, dict):
        if "$link" in val and len(val) == 1:
            return CID.decode(val["$link"])

        if "$bytes" in val and len(val) == 1:
            return base64.b64decode(val["$bytes"])

        return {key: json_to_ipld(val[key]) for key in val.keys()}

    return val


def ipld_to_json(val: IpldValue) -> JsonValue:
    if isinstance(val, list):
        return [ipld_to_json(item) for item in val]

    if val:
        if isinstance(val, bytes):
            return {"$bytes": base64.b64encode(val).decode("utf-8")}

        if (cid := as_cid(val)) is not None:
            return {"$link": str(cid)}

        if isinstance(val, dict):
            return {key: ipld_to_json(val[key]) for key in val.keys()}

    return val


def ipld_equals(a: IpldValue, b: IpldValue) -> bool:
    if isinstance(a, list) and isinstance(b, list):
        if len(a) != len(b):
            return False

        for i in range(len(a)):
            if not ipld_equals(a[i], b[i]):
                return False

        return True

    if a and b:
        if isinstance(a, bytes) and isinstance(b, bytes):
            return a == b

        if (a_cid := as_cid(a)) and (b_cid := as_cid(b)):
            # TODO: what is correct equality
            return a_cid == b_cid

        if isinstance(a, dict) and isinstance(b, dict):
            if len(a.keys()) != len(b.keys()):
                return False

            for key in a.keys():
                if not ipld_equals(a[key], b[key]):
                    return False
            return True

    return a == b



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.api.lexicons.app.bsky.embed.record import View, Record
from psychonaut.api.lexicons.app.bsky.embed.images import View
from psychonaut.api.lexicons.app.bsky.embed.external import View


class RecordWithMedia(BaseModel):
    """
    [none provided by spec]
    """

    record: Record
    media: Any


class View(BaseModel):
    """
    [none provided by spec]
    """

    record: View
    media: Any



from psychonaut.cli.group import cli

# Imports to assemble the cli
from psychonaut.cli.config import * # noqa
from psychonaut.cli.stream import * # noqa
from psychonaut.cli.useful_queries import * # noqa
from psychonaut.cli.poasting import * # noqa
from psychonaut.cli.graph import * # noqa




# TODO: test poetry install
if __name__ == "__main__":
    cli()



import json
from typing import List, Tuple
from .types import LexXrpcBody, LexiconDoc, LexObject
from pathlib import Path


ROOT_DIR = Path(__file__).parent.parent.parent
LEXICONS_DIR = ROOT_DIR / "lexicons"


class TestProcedure:

    def test_lexicon_input(self):
        # Smoketest
        json_doc = _load_lexicon_doc("com", "atproto", "server", "createAccount")
        input_doc = json_doc["defs"]["main"]["input"]
        got = LexXrpcBody(**input_doc)

        assert got.encoding == "application/json"

        # schema_kludge is a LexObject
        assert isinstance(got.schema_kludge, LexObject)

        schema = got.schema_kludge
        assert schema.type == "object"
        assert schema.required == ["handle", "email", "password"]
        assert len(schema.properties) == 5

        email = schema.properties["email"]
        assert email.type == "string"

        handle = schema.properties["handle"]
        assert handle.type == "string"
        assert handle.format == "handle"

        invite_code = schema.properties["inviteCode"]
        assert invite_code.type == "string"

        password = schema.properties["password"]
        assert password.type == "string"

        recovery_key = schema.properties["recoveryKey"]
        assert recovery_key.type == "string"


    def test_lexicon_output(self):
        # Smoketest

        json_doc = _load_lexicon_doc("com", "atproto", "server", "createAccount")
        output_doc = json_doc["defs"]["main"]["output"]
        got = LexXrpcBody(**output_doc)

        assert got.encoding == "application/json"

        # schema_kludge is a LexObject
        assert isinstance(got.schema_kludge, LexObject)

        schema = got.schema_kludge
        assert schema.type == "object"
        assert schema.required == ["accessJwt", "refreshJwt", "handle", "did"]

        access_jwt = schema.properties["accessJwt"]
        assert access_jwt.type == "string"

        refresh_jwt = schema.properties["refreshJwt"]
        assert refresh_jwt.type == "string"

        handle = schema.properties["handle"]
        assert handle.type == "string"
        assert handle.format == "handle"

        did = schema.properties["did"]
        assert did.type == "string"


    def test_lexicon_doc_from_dict(self):
        json_doc = _load_lexicon_doc("com", "atproto", "server", "createAccount")
        doc = LexiconDoc(**json_doc)
        assert doc.lexicon == 1
        assert doc.id == "com.atproto.server.createAccount"
        assert doc.revision is None
        assert doc.description is None
        assert len(doc.defs) == 1

        # the type of the "main" def is "procedure"
        main = doc.defs["main"]
        assert main.type == "procedure"
        assert main.description == "Create an account."

        input = main.input
        assert input.encoding == "application/json"
        schema = input.schema_kludge
        assert schema.type == "object"
        assert schema.required == ["handle", "email", "password"]

        output = main.output
        assert output.encoding == "application/json"
        schema = output.schema_kludge
        assert schema.type == "object"
        assert schema.required == ["accessJwt", "refreshJwt", "handle", "did"]

        assert len(main.errors) == 5
        assert main.errors[0].name == "InvalidHandle"
        assert main.errors[1].name == "InvalidPassword"
        assert main.errors[2].name == "InvalidInviteCode"
        assert main.errors[3].name == "HandleNotAvailable"
        assert main.errors[4].name == "UnsupportedDomain"



def _load_lexicon_doc(*parts: Tuple[str]) -> LexiconDoc:
    parts = list(parts)
    parts[-1] = parts[-1] + ".json"
    full_path = LEXICONS_DIR.joinpath(*parts)
    with open(full_path, "r") as f:
        return json.load(f)



import base64
from contextlib import contextmanager
import json
from carbox.car import read_car

from psychonaut.firehose.serde import (
    _json_encode_kludge,
    read_event_pair,
)
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Tuple


@contextmanager
def noop_segment_writer():
    yield lambda msg, first_block: None


def write_length_prefixed_msg(msg: bytes, fp) -> int:
    length = len(msg)
    length_bytes = length.to_bytes(4, "big")
    fp.write(length_bytes)
    fp.write(msg)
    return 4 + length


FINALIZE_SEGMENT = Callable[[Path, int], None]


def noop_finalize_segment(path: Path, bytes_written: int):
    pass


class LengthDelimitedStreamSegmentWriter:
    def __init__(
        self,
        output_dir: Path | str,
        segment_size: int = 100_000_000,
        finalize_segment: FINALIZE_SEGMENT = noop_finalize_segment,
    ):
        self._output_dir = Path(output_dir)
        self._output_dir.mkdir(exist_ok=True, parents=True)
        self._segment_size = segment_size

        self._segment_path = None
        self._segment_fp = None
        self._bytes_written = 0
        self._finalize_segment = finalize_segment

        self._init_next_segment()

    def _init_next_segment(self):
        if self._segment_fp:
            self._segment_fp.close()
            self._finalize_segment(self._segment_path, self._bytes_written)

        now = datetime.now()
        dir_name = now.strftime("%Y-%m-%d")
        file_name = now.strftime("%H-%M-%S.length-prefixed")
        segment_dir = self._output_dir / dir_name
        if not segment_dir.exists():
            segment_dir.mkdir(exist_ok=True, parents=True)

        self._segment_path = segment_dir / file_name
        self._segment_fp = self._segment_path.open("wb")
        self._bytes_written = 0

    @property
    def segment_path(self):
        return self._segment_path

    def write(self, data: bytes):
        self._bytes_written += write_length_prefixed_msg(data, self._segment_fp)

        if self._bytes_written >= self._segment_size:
            self._init_next_segment()

    def __call__(self, msg: bytes, *args, **kwargs):
        self.write(msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._segment_fp.close()


def b64line_iter(stream_path: Path):
    with stream_path.open() as f:
        for line in f:
            yield base64.b64decode(line)


def b64line_cbor2_iter(stream_path: Path):
    for msg in b64line_iter(stream_path):
        yield read_event_pair(msg)


def length_prefixed_iter(stream_path: Path):
    # I think the limit is 1_000_000 bytes and most are much smaller than that
    # but i don't have an optimized uvarint reader yet so just do 4 bytes.
    with stream_path.open("rb") as f:
        while True:
            length_bytes = f.read(4)
            if not length_bytes:
                break
            length = int.from_bytes(length_bytes, "big")
            msg = f.read(length)
            yield msg


def convert_b64_to_length_prefixed(
    input_file: Path, output_dir: Path
) -> Tuple[Path, int]:
    n_bytes = 0

    output_name = input_file.name.replace(".b64-lines", ".length-prefixed")
    output_file = output_dir / output_name

    with output_file.open("wb") as out_fp:
        for msg in b64line_iter(input_file):
            n_bytes += write_length_prefixed_msg(msg, out_fp)

    return output_file, n_bytes


def convert_b64_to_length_prefixed_all(
    input_dir: Path, output_dir: Path, verbose=False, skip_empty=False
):
    for input_file in sorted(input_dir.glob("*.b64-lines")):
        input_mb = input_file.stat().st_size / 1_000_000
        if input_mb == 1 and not skip_empty:
            continue

        _, output_bytes = convert_b64_to_length_prefixed(input_file, output_dir)

        output_mb = output_bytes / 1_000_000
        if verbose:
            print(f"{input_file.name}: {input_mb:.2f}MB -> {output_mb:.2f}MB")


def stream_to_stdout(length_prefixed_file: Path):
    for msg in length_prefixed_iter(length_prefixed_file):
        header, event = read_event_pair(msg)
        roots, blocks = read_car(event["blocks"])
        first_block = blocks[0]

        if first_block:
            # TODO: finish pydantic models for records so no kludge needed
            try:
                event["first_block_cid"] = first_block.cid
                event["first_block_msg"] = first_block.decoded
                del event["blocks"]
                line = json.dumps(_json_encode_kludge(event))
            except TypeError:
                print(first_block)
                raise
            print(line)


def iter_length_prefixed_paths(data_dir: Path, verbose=False):
    assert data_dir.exists(), f"{data_dir} does not exist"
    
    for file_path in sorted(data_dir.glob("**/*.length-prefixed")):
        mb_size = file_path.stat().st_size / 1_000_000
        path_rel_data_dir = file_path.relative_to(data_dir)
        if verbose:
            print(f"{path_rel_data_dir}: {mb_size} mb")

        yield file_path


def read_all_length_prefixed_bytes(data_dir: Path, verbose=False):
    for file_path in iter_length_prefixed_paths(data_dir, verbose):
        for msg in length_prefixed_iter(file_path):
            yield msg


from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid


class GetRecordReq(BaseModel):
    """
    View details about a record.
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.getRecord"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



import pytest
import os
import json


@pytest.fixture
def load_test_fixture(request):
    def _load_file(filename, as_json=False, fixtures_dir=True):
        test_module_path = os.path.dirname(request.fspath)
        file_relative_path = filename
        if fixtures_dir:
            file_relative_path = os.path.join("fixtures", filename)
        file_absolute_path = os.path.join(test_module_path, file_relative_path)
        with open(file_absolute_path, "r") as file:
            res = file.read()
            if as_json:
                res = json.loads(res)
        return res
    return _load_file


from typing import Any, Optional
from pydantic import BaseModel, Field


class Profile(BaseModel):
    """
    [none provided by spec]
    """

    displayName: Optional[str] = Field(default=None, max_length=640)
    description: Optional[str] = Field(default=None, max_length=2560)
    avatar: Optional[Any] = None
    banner: Optional[Any] = None

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.actor.profile"



from multiformats import CID
from .events import IndexRecord


def test_index_record():
    r = IndexRecord(
        type="index_record",
        action="Create",
        uri="at://foo/bar",
        cid=CID.decode("zdpuArKcqh4Bfc5ufSWKTSS1jFRYJ47gpuxCEVXeWdMEjDpAM"),
        timestamp="2021-01-01T00:00:00Z",
    )



from psychonaut.lexicon.ctx import GenCtx
from .types import LexArray, LexInteger, LexBoolean, LexString
from .fields import _generate_array_field, _generate_integer_field, _generate_bool_field, _generate_str_field


def test_generate_integer_field():
    opts = {
        "type": "integer",
        "default": 50,
        "minimum": 1,
        "maximum": 100,
    }

    example = LexInteger(**opts)
    s = _generate_integer_field("test_k", True, example)
    assert s == "test_k: int = Field(default=50, ge=1, le=100)"
    s = _generate_integer_field("test_k", False, example)
    assert s == "test_k: Optional[int] = Field(default=50, ge=1, le=100)"

    opts["description"] = "a description"
    example = LexInteger(**opts)
    s = _generate_integer_field("test_k", True, example)
    expected = (
        "test_k: int = Field(default=50, description='a description', ge=1, le=100)"
    )
    assert s == expected


def test_generate_bool_field():
    opts = {"type": "boolean"}

    example = LexBoolean(**opts)
    s = _generate_bool_field("test_k", True, example)
    assert s == "test_k: bool = Field(...)"

    opts["default"] = True
    example = LexBoolean(**opts)
    s = _generate_bool_field("test_k", True, example)
    assert s == "test_k: bool = Field(default=True)"


def test_generate_str_field():
    opts = {"type": "string", "description": "The DID of the repo."}

    ctx, example = GenCtx(), LexString(**opts)
    s = _generate_str_field(ctx, "test_k", True, example)
    assert s == "test_k: str = Field(..., description='The DID of the repo.')"

    opts["knownValues"] = ["recent", "usage"]
    del opts["description"]
    ctx, example = GenCtx(), LexString(**opts)
    s = _generate_str_field(ctx, "test_k", True, example)
    assert s == "test_k: str = Field(..., known_values=['recent', 'usage'])"

    opts["format"] = "cid"
    del opts["knownValues"]
    ctx, example = GenCtx(), LexString(**opts)
    s = _generate_str_field(ctx, "test_k", True, example)
    assert s == "test_k: str = Field(..., pre=True, validator=validate_cid)"

    opts["format"] = "did"
    ctx, example = GenCtx(), LexString(**opts)
    s = _generate_str_field(ctx, "test_k", True, example)
    assert s == "test_k: str = Field(..., pre=True, validator=validate_did)"

    opts["format"] = "handle"
    ctx, example = GenCtx(), LexString(**opts)
    s = _generate_str_field(ctx, "test_k", True, example)
    assert s == "test_k: str = Field(..., pre=True, validator=validate_handle)"


def test_generate_array_field():
    opts = {
        "type": "array",
        "description": "a desc",
        "items": {
            "type": "string",
            "format": "did",
        },
    }
    ctx, example = GenCtx(), LexArray(**opts)
    s = _generate_array_field(ctx, "test_k", True, example)
    assert s == "test_k: List[str] = Field(..., description='a desc', pre=True, validator=validate_array(validate_did))"



# repo/src/block-map.ts
# from typing import Dict, List, Tuple, Union
# from multiformats import CID
# from pydantic import BaseModel
# #from placeholder_lexicon import lex_to_ipld, LexValue
# from placeholder_common import data_to_cbor_block


# class Entry(BaseModel):
#     cid: CID
#     bytes: bytes


# class BlockMap:
#     def __init__(self):
#         self.map: Dict[str, bytes] = {}

#     async def add(self, value: LexValue) -> CID:
#         block = await data_to_cbor_block(lex_to_ipld(value))
#         self.set(block.cid, block.bytes)
#         return block.cid

#     def set(self, cid: CID, bytes: bytes) -> None:
#         self.map[cid.to_string()] = bytes

#     def get(self, cid: CID) -> Union[bytes, None]:
#         return self.map.get(cid.to_string())

#     def delete(self, cid: CID) -> None:
#         del self.map[cid.to_string()]

#     def get_many(self, cids: List[CID]) -> Tuple["BlockMap", List[CID]]:
#         missing: List[CID] = []
#         blocks = BlockMap()
#         for cid in cids:
#             got = self.map.get(cid.to_string())
#             if got:
#                 blocks.set(cid, got)
#             else:
#                 missing.append(cid)
#         return blocks, missing

#     def has(self, cid: CID) -> bool:
#         return cid.to_string() in self.map

#     def clear(self) -> None:
#         self.map.clear()

#     def for_each(self, cb) -> None:
#         for key, val in self.map.items():
#             cb(val, CID.parse(key))

#     def entries(self) -> List[Entry]:
#         entries: List[Entry] = []
#         self.for_each(lambda bytes, cid: entries.append(Entry(cid=cid, bytes=bytes)))
#         return entries

#     def cids(self) -> List[CID]:
#         return [entry.cid for entry in self.entries()]

#     def add_map(self, to_add: "BlockMap") -> None:
#         to_add.for_each(lambda bytes, cid: self.set(cid, bytes))

#     def size(self) -> int:
#         return len(self.map)

#     def byte_size(self) -> int:
#         size = 0
#         self.for_each(lambda bytes: size.__add__(len(bytes)))
#         return size

#     def equals(self, other: "BlockMap") -> bool:
#         if self.size() != other.size():
#             return False
#         for entry in self.entries():
#             other_bytes = other.get(entry.cid)
#             if not other_bytes:
#                 return False
#             if entry.bytes != other_bytes:
#                 return False
#         return True



import pytest
from psychonaut.uri import AtUri, parse, parse_relative
from psychonaut.uri.validation import ensure_valid_at_uri, ensure_valid_at_uri_regex

URI_TESTS = [
    ["foo.com", "foo.com", "", "", ""],
    ["at://foo.com", "foo.com", "", "", ""],
    ["at://foo.com/", "foo.com", "/", "", ""],
    ["at://foo.com/foo", "foo.com", "/foo", "", ""],
    ["at://foo.com/foo/", "foo.com", "/foo/", "", ""],
    ["at://foo.com/foo/bar", "foo.com", "/foo/bar", "", ""],
    ["at://foo.com?foo=bar", "foo.com", "", "foo=bar", ""],
    ["at://foo.com?foo=bar&baz=buux", "foo.com", "", "foo=bar&baz=buux", ""],
    ["at://foo.com/?foo=bar", "foo.com", "/", "foo=bar", ""],
    ["at://foo.com/foo?foo=bar", "foo.com", "/foo", "foo=bar", ""],
    ["at://foo.com/foo/?foo=bar", "foo.com", "/foo/", "foo=bar", ""],
    ["at://foo.com#hash", "foo.com", "", "", "#hash"],
    ["at://foo.com/#hash", "foo.com", "/", "", "#hash"],
    ["at://foo.com/foo#hash", "foo.com", "/foo", "", "#hash"],
    ["at://foo.com/foo/#hash", "foo.com", "/foo/", "", "#hash"],
    ["at://foo.com?foo=bar#hash", "foo.com", "", "foo=bar", "#hash"],
    [
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo/",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo/",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo/bar",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo/bar",
        "",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw?foo=bar",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "foo=bar",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw?foo=bar&baz=buux",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "foo=bar&baz=buux",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/?foo=bar",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/",
        "foo=bar",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo?foo=bar",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo",
        "foo=bar",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo/?foo=bar",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo/",
        "foo=bar",
        "",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw#hash",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "",
        "#hash",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/#hash",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/",
        "",
        "#hash",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo#hash",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo",
        "",
        "#hash",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw/foo/#hash",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "/foo/",
        "",
        "#hash",
    ],
    [
        "at://did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw?foo=bar#hash",
        "did:example:EiAnKD8-jfdd0MDcZUjAbRgaThBrMxPTFOxcnfJhI7Ukaw",
        "",
        "foo=bar",
        "#hash",
    ],
    ["did:web:localhost%3A1234", "did:web:localhost%3A1234", "", "", ""],
    ["at://did:web:localhost%3A1234", "did:web:localhost%3A1234", "", "", ""],
    [
        "at://did:web:localhost%3A1234/",
        "did:web:localhost%3A1234",
        "/",
        "",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/foo",
        "did:web:localhost%3A1234",
        "/foo",
        "",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/foo/",
        "did:web:localhost%3A1234",
        "/foo/",
        "",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/foo/bar",
        "did:web:localhost%3A1234",
        "/foo/bar",
        "",
        "",
    ],
    [
        "at://did:web:localhost%3A1234?foo=bar",
        "did:web:localhost%3A1234",
        "",
        "foo=bar",
        "",
    ],
    [
        "at://did:web:localhost%3A1234?foo=bar&baz=buux",
        "did:web:localhost%3A1234",
        "",
        "foo=bar&baz=buux",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/?foo=bar",
        "did:web:localhost%3A1234",
        "/",
        "foo=bar",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/foo?foo=bar",
        "did:web:localhost%3A1234",
        "/foo",
        "foo=bar",
        "",
    ],
    [
        "at://did:web:localhost%3A1234/foo/?foo=bar",
        "did:web:localhost%3A1234",
        "/foo/",
        "foo=bar",
        "",
    ],
    [
        "at://did:web:localhost%3A1234#hash",
        "did:web:localhost%3A1234",
        "",
        "",
        "#hash",
    ],
    [
        "at://did:web:localhost%3A1234/#hash",
        "did:web:localhost%3A1234",
        "/",
        "",
        "#hash",
    ],
    [
        "at://did:web:localhost%3A1234/foo#hash",
        "did:web:localhost%3A1234",
        "/foo",
        "",
        "#hash",
    ],
    [
        "at://did:web:localhost%3A1234/foo/#hash",
        "did:web:localhost%3A1234",
        "/foo/",
        "",
        "#hash",
    ],
    [
        "at://did:web:localhost%3A1234?foo=bar#hash",
        "did:web:localhost%3A1234",
        "",
        "foo=bar",
        "#hash",
    ],
    [
        "at://4513echo.bsky.social/app.bsky.feed.post/3jsrpdyf6ss23",
        "4513echo.bsky.social",
        "/app.bsky.feed.post/3jsrpdyf6ss23",
        "",
        "",
    ],
]
# TODO: I should make all of the tests as json and make a suite for anyone


class TestAtURLS:
    @pytest.mark.parametrize("uri,hostname,pathname,search,hash", URI_TESTS)
    def test_parse(self, uri, hostname, pathname, search, hash):
        urip = AtUri(uri)
        assert urip.protocol == "at:"
        assert urip.host == hostname
        assert urip.hostname == hostname
        assert urip.origin == f"at://{hostname}"
        assert urip.pathname == pathname
        assert urip.search == search, urip.search_params
        assert urip.hash == hash

    @pytest.mark.parametrize(
        "uri,collection,rkey",
        [
            ("at://foo.com", "", ""),
            ("at://foo.com/com.example.foo", "com.example.foo", ""),
            ("at://foo.com/com.example.foo/123", "com.example.foo", "123"),
        ],
    )
    def test_handles_atp_specific_parsing(self, uri, collection, rkey):
        urip = AtUri(uri)
        assert urip.collection == collection
        assert urip.rkey == rkey

    def test_supports_modifications(self):
        urip = AtUri("at://foo.com")
        assert str(urip) == "at://foo.com/"

        urip.host = "bar.com"
        assert str(urip) == "at://bar.com/"
        urip.host = "did:web:localhost%3A1234"
        assert str(urip) == "at://did:web:localhost%3A1234/"
        urip.host = "foo.com"

        urip.pathname = "/"
        assert str(urip) == "at://foo.com/"
        urip.pathname = "/foo"
        assert str(urip) == "at://foo.com/foo"
        urip.pathname = "foo"
        assert str(urip) == "at://foo.com/foo"

        urip.collection = "com.example.foo"
        urip.rkey = "123"
        assert str(urip) == "at://foo.com/com.example.foo/123"
        urip.rkey = "124"
        assert str(urip) == "at://foo.com/com.example.foo/124"
        urip.collection = "com.other.foo"
        assert str(urip) == "at://foo.com/com.other.foo/124"
        urip.pathname = ""
        urip.rkey = "123"
        assert str(urip) == "at://foo.com/undefined/123"
        urip.pathname = "foo"

        urip.search = "?foo=bar"
        assert str(urip) == "at://foo.com/foo?foo=bar"
        urip.search_params["baz"] = ["buux"]
        assert str(urip) == "at://foo.com/foo?foo=bar&baz=buux"

        urip.hash = "#hash"
        assert str(urip) == "at://foo.com/foo?foo=bar&baz=buux#hash"
        urip.hash = "hash"
        assert str(urip) == "at://foo.com/foo?foo=bar&baz=buux#hash"

    @pytest.mark.parametrize(
        "base",
        [
            "did:web:localhost%3A1234",
            "at://did:web:localhost%3A1234",
            "at://did:web:localhost%3A1234/foo/bar?foo=bar&baz=buux#hash",
            "did:web:localhost%3A1234",
            "at://did:web:localhost%3A1234",
            "at://did:web:localhost%3A1234/foo/bar?foo=bar&baz=buux#hash",
        ],
    )
    def test_supports_relative_uris(self, base):
        tests = [
            ("", "", "", ""),
            ("/", "/", "", ""),
            ("/foo", "/foo", "", ""),
            ("/foo/", "/foo/", "", ""),
            ("/foo/bar", "/foo/bar", "", ""),
            ("?foo=bar", "", "foo=bar", ""),
            ("?foo=bar&baz=buux", "", "foo=bar&baz=buux", ""),
            ("/?foo=bar", "/", "foo=bar", ""),
            ("/foo?foo=bar", "/foo", "foo=bar", ""),
            ("/foo/?foo=bar", "/foo/", "foo=bar", ""),
            ("#hash", "", "", "#hash"),
            ("/#hash", "/", "", "#hash"),
            ("/foo#hash", "/foo", "", "#hash"),
            ("/foo/#hash", "/foo/", "", "#hash"),
            ("?foo=bar#hash", "", "foo=bar", "#hash"),
        ]
        basep = AtUri(base)
        for relative, pathname, search, hash in tests:
            urip = AtUri(relative, base)
            assert urip.protocol == "at:"
            assert urip.host == basep.host
            assert urip.hostname == basep.hostname
            assert urip.origin == basep.origin
            assert urip.pathname == pathname
            assert urip.search == search
            assert urip.hash == hash


# at url ======================================================================


def expect_valid_at(h):
    ensure_valid_at_uri(h)
    ensure_valid_at_uri_regex(h)


def expect_invalid_at(h):
    with pytest.raises(Exception):
        ensure_valid_at_uri(h)
    with pytest.raises(Exception):
        ensure_valid_at_uri_regex(h)


class TestAtUrlValidation:
    def test_spec_basics(self):
        expect_valid_at("at://did:plc:asdf123")
        expect_valid_at("at://user.bsky.social")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/record")

        expect_valid_at("at://did:plc:asdf123#/frag")
        expect_valid_at("at://user.bsky.social#/frag")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post#/frag")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/record#/frag")

        expect_invalid_at("a://did:plc:asdf123")
        expect_invalid_at("at//did:plc:asdf123")
        expect_invalid_at("at:/a/did:plc:asdf123")
        expect_invalid_at("at:/did:plc:asdf123")
        expect_invalid_at("AT://did:plc:asdf123")
        expect_invalid_at("http://did:plc:asdf123")
        expect_invalid_at("://did:plc:asdf123")
        expect_invalid_at("at:did:plc:asdf123")
        expect_invalid_at("at:/did:plc:asdf123")
        expect_invalid_at("at:///did:plc:asdf123")
        expect_invalid_at("at://:/did:plc:asdf123")
        expect_invalid_at("at:/ /did:plc:asdf123")
        expect_invalid_at("at://did:plc:asdf123 ")
        expect_invalid_at("at://did:plc:asdf123/ ")
        expect_invalid_at(" at://did:plc:asdf123")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post ")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post# ")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post#/ ")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post#/frag ")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post#fr ag")
        expect_invalid_at("//did:plc:asdf123")
        expect_invalid_at("at://name")
        expect_invalid_at("at://name.0")
        expect_invalid_at("at://diD:plc:asdf123")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.p@st")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.p$st")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.p%st")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.p&st")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.p()t")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed_post")
        expect_invalid_at("at://did:plc:asdf123/-com.atproto.feed.post")
        expect_invalid_at("at://did:plc:asdf@123/com.atproto.feed.post")

        expect_invalid_at("at://DID:plc:asdf123")
        expect_invalid_at("at://user.bsky.123")
        expect_invalid_at("at://bsky")
        expect_invalid_at("at://did:plc:")
        expect_invalid_at("at://did:plc:")
        expect_invalid_at("at://frag")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/" + "o" * 800)
        expect_invalid_at(
            "at://did:plc:asdf123/com.atproto.feed.post/" + "o" * 8200,
        )

    def test_edge_cases(self):
        expect_invalid_at("at://user.bsky.social//")
        expect_invalid_at("at://user.bsky.social//com.atproto.feed.post")

        expect_invalid_at("at://user.bsky.social/com.atproto.feed.post//")
        expect_invalid_at(
            "at://did:plc:asdf123/com.atproto.feed.post/asdf123/more/more"
        )

        expect_invalid_at("at://did:plc:asdf123/short/stuff")
        expect_invalid_at("at://did:plc:asdf123/12345")

    def test_no_trailing_slashes(self):
        expect_valid_at("at://did:plc:asdf123")
        expect_invalid_at("at://did:plc:asdf123/")

        expect_valid_at("at://user.bsky.social")
        expect_invalid_at("at://user.bsky.social/")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post/")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/record")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post/record/")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post/record/#/frag")

    def test_strict_paths(self):
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/asdf123")
        expect_invalid_at("at://did:plc:asdf123/com.atproto.feed.post/asdf123/asdf")

    def test_permissive_record_keys(self):
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/asdf123")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/a")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%23")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/$@!*)(:,;~.sdf123")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/~'sdf123")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/$")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/@")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/!")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/*")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/(")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/,")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/;")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/abc%30123")

    def test_url_encoding(self):
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%30")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%3")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%zz")
        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post/%%%")

    def test_fragments(self):
        expect_valid_at("at://did:plc:asdf123#/frac")

        expect_invalid_at("at://did:plc:asdf123#")
        expect_invalid_at("at://did:plc:asdf123##")
        expect_invalid_at("#at://did:plc:asdf123")
        expect_invalid_at("at://did:plc:asdf123#/asdf#/asdf")

        expect_valid_at("at://did:plc:asdf123#/com.atproto.feed.post")
        expect_valid_at("at://did:plc:asdf123#/com.atproto.feed.post/")
        expect_valid_at("at://did:plc:asdf123#/asdf/")

        expect_valid_at("at://did:plc:asdf123/com.atproto.feed.post#/$@!*():,;~.sdf123")
        expect_valid_at("at://did:plc:asdf123#/[asfd]")

        expect_valid_at("at://did:plc:asdf123#/$")
        expect_valid_at("at://did:plc:asdf123#/*")
        expect_valid_at("at://did:plc:asdf123#/;")
        expect_valid_at("at://did:plc:asdf123#/,")



class FirehoseExponentialBackoff:

    def __init__(self, initial_wait: int = 0, base: int = 2, max_wait: int = 60):
        self._initial_wait = initial_wait
        self._base = base
        self._max_wait = max_wait
        self._failures = 0

    def reset(self):
        self._failures = 0

    def next_sleep_time(self) -> int:
        sleep_time = self._initial_wait

        if self._failures > 0:
            sleep_time = min(self._max_wait, self._base ** self._failures)
        
        self._failures += 1

        return sleep_time
            


"""
See:

https://github.dev/bluesky-social/atproto/blob/main/packages/lexicon/src/validators/formats.ts
"""

ATP_SPECIFIC = [
    "at",
    "atp",
    "plc",
    "pds",
    "did",
    "repo",
    "tid",
    "nsid",
    "xrpc",
    "lex",
    "lexicon",
    "bsky",
    "bluesky",
    "handle",
]

# naively pulled from: https://radwebhosting.com/client_area/knowledgebase/196/List-of-Banned-Subdomain-Prefixes.html
COMMONLY_RESERVED = [
    "about",
    "abuse",
    "access",
    "account",
    "accounts",
    "acme",
    "activate",
    "activities",
    "activity",
    "ad",
    "add",
    "address",
    "adm",
    "admanager",
    "admin",
    "administration",
    "administrator",
    "administrators",
    "admins",
    "ads",
    "adsense",
    "adult",
    "advertising",
    "adwords",
    "affiliate",
    "affiliatepage",
    "affiliates",
    "afp",
    "ajax",
    "all",
    "alpha",
    "analysis",
    "analytics",
    "android",
    "anon",
    "anonymous",
    "answer",
    "answers",
    "ap",
    "api",
    "apis",
    "app",
    "appengine",
    "appnews",
    "apps",
    "archive",
    "archives",
    "article",
    "asdf",
    "asset",
    "assets",
    "auth",
    "authentication",
    "avatar",
    "backup",
    "bank",
    "banner",
    "banners",
    "base",
    "beginners",
    "beta",
    "billing",
    "bin",
    "binaries",
    "binary",
    "blackberry",
    "blog",
    "blogs",
    "blogsearch",
    "board",
    "book",
    "bookmark",
    "bookmarks",
    "books",
    "bot",
    "bots",
    "bug",
    "bugs",
    "business",
    "buy",
    "buzz",
    "cache",
    "calendar",
    "call",
    "campaign",
    "cancel",
    "captcha",
    "career",
    "careers",
    "cart",
    "catalog",
    "catalogs",
    "categories",
    "category",
    "cdn",
    "cgi",
    "cgi-bin",
    "changelog",
    "chart",
    "charts",
    "chat",
    "check",
    "checked",
    "checking",
    "checkout",
    "client",
    "cliente",
    "clients",
    "clients1",
    "cnarne",
    "code",
    "comercial",
    "comment",
    "comments",
    "communities",
    "community",
    "company",
    "compare",
    "compras",
    "config",
    "configuration",
    "confirm",
    "confirmation",
    "connect",
    "contact",
    "contacts",
    "contactus",
    "contact-us",
    "contact_us",
    "content",
    "contest",
    "contribute",
    "contributor",
    "contributors",
    "coppa",
    "copyright",
    "copyrights",
    "core",
    "corp",
    "countries",
    "country",
    "cpanel",
    "create",
    "css",
    "cssproxy",
    "customise",
    "customize",
    "dashboard",
    "data",
    "db",
    "default",
    "delete",
    "demo",
    "design",
    "designer",
    "desktop",
    "destroy",
    "dev",
    "devel",
    "developer",
    "developers",
    "devs",
    "diagram",
    "diary",
    "dict",
    "dictionary",
    "die",
    "dir",
    "directory",
    "direct_messages",
    "direct-messages",
    "dist",
    "diversity",
    "dl",
    "dmca",
    "doc",
    "docs",
    "documentation",
    "documentations",
    "documents",
    "domain",
    "domains",
    "donate",
    "download",
    "downloads",
    "e",
    "e-mail",
    "earth",
    "ecommerce",
    "edit",
    "edits",
    "editor",
    "edu",
    "education",
    "email",
    "embed",
    "embedded",
    "employment",
    "employments",
    "empty",
    "enable",
    "encrypted",
    "end",
    "engine",
    "enterprise",
    "enterprises",
    "entries",
    "entry",
    "error",
    "errorlog",
    "errors",
    "eval",
    "event",
    "example",
    "examplecommunity",
    "exampleopenid",
    "examplesyn",
    "examplesyndicated",
    "exampleusername",
    "exchange",
    "exit",
    "explore",
    "faq",
    "faqs",
    "favorite",
    "favorites",
    "favourite",
    "favourites",
    "feature",
    "features",
    "feed",
    "feedback",
    "feedburner",
    "feedproxy",
    "feeds",
    "file",
    "files",
    "finance",
    "folder",
    "folders",
    "first",
    "following",
    "forgot",
    "form",
    "forms",
    "forum",
    "forums",
    "founder",
    "free",
    "friend",
    "friends",
    "ftp",
    "fuck",
    "fun",
    "fusion",
    "gadget",
    "gadgets",
    "game",
    "games",
    "gears",
    "general",
    "geographic",
    "get",
    "gettingstarted",
    "gift",
    "gifts",
    "gist",
    "git",
    "github",
    "gmail",
    "go",
    "golang",
    "goto",
    "gov",
    "graph",
    "graphs",
    "group",
    "groups",
    "guest",
    "guests",
    "guide",
    "guides",
    "hack",
    "hacks",
    "head",
    "help",
    "home",
    "homepage",
    "host",
    "hosting",
    "hostmaster",
    "hostname",
    "howto",
    "how-to",
    "how_to",
    "html",
    "htrnl",
    "http",
    "httpd",
    "https",
    "i",
    "iamges",
    "icon",
    "icons",
    "id",
    "idea",
    "ideas",
    "im",
    "image",
    "images",
    "img",
    "imap",
    "inbox",
    "inboxes",
    "index",
    "indexes",
    "info",
    "information",
    "inquiry",
    "intranet",
    "investor",
    "investors",
    "invitation",
    "invitations",
    "invite",
    "invoice",
    "invoices",
    "imac",
    "ios",
    "ipad",
    "iphone",
    "irc",
    "irnages",
    "irng",
    "is",
    "issue",
    "issues",
    "it",
    "item",
    "items",
    "java",
    "javascript",
    "job",
    "jobs",
    "join",
    "js",
    "json",
    "jump",
    "kb",
    "knowledge-base",
    "knowledgebase",
    "lab",
    "labs",
    "language",
    "languages",
    "last",
    "ldap_status",
    "ldap-status",
    "ldapstatus",
    "legal",
    "license",
    "licenses",
    "link",
    "links",
    "linux",
    "list",
    "lists",
    "livejournal",
    "lj",
    "local",
    "locale",
    "location",
    "log",
    "log-in",
    "log-out",
    "login",
    "logout",
    "logs",
    "log_in",
    "log_out",
    "m",
    "mac",
    "macos",
    "macosx",
    "mac-os",
    "mac-os-x",
    "mac_os_x",
    "mail",
    "mailer",
    "mailing",
    "main",
    "maintenance",
    "manage",
    "manager",
    "manual",
    "map",
    "maps",
    "marketing",
    "master",
    "me",
    "media",
    "member",
    "members",
    "memories",
    "memory",
    "merchandise",
    "message",
    "messages",
    "messenger",
    "mg",
    "microblog",
    "microblogs",
    "mine",
    "mis",
    "misc",
    "mms",
    "mob",
    "mobile",
    "model",
    "models",
    "money",
    "movie",
    "movies",
    "mp3",
    "mp4",
    "msg",
    "msn",
    "music",
    "mx",
    "my",
    "mymme",
    "mysql",
    "name",
    "named",
    "nan",
    "navi",
    "navigation",
    "net",
    "network",
    "networks",
    "new",
    "news",
    "newsletter",
    "nick",
    "nickname",
    "nil",
    "none",
    "notes",
    "noticias",
    "notification",
    "notifications",
    "notify",
    "ns",
    "ns1",
    "ns2",
    "ns3",
    "ns4",
    "ns5",
    "null",
    "oauth",
    "oauth-clients",
    "oauth_clients",
    "ocsp",
    "offer",
    "offers",
    "official",
    "old",
    "online",
    "openid",
    "operator",
    "option",
    "options",
    "order",
    "orders",
    "org",
    "organization",
    "organizations",
    "other",
    "overview",
    "owner",
    "owners",
    "p0rn",
    "pack",
    "page",
    "pager",
    "pages",
    "paid",
    "panel",
    "partner",
    "partnerpage",
    "partners",
    "password",
    "patch",
    "pay",
    "payment",
    "people",
    "perl",
    "person",
    "phone",
    "photo",
    "photoalbum",
    "photos",
    "php",
    "phpmyadmin",
    "phppgadmin",
    "phpredisadmin",
    "pic",
    "pics",
    "picture",
    "pictures",
    "ping",
    "pixel",
    "places",
    "plan",
    "plans",
    "plugin",
    "plugins",
    "podcasts",
    "policies",
    "policy",
    "pop",
    "pop3",
    "popular",
    "porn",
    "portal",
    "portals",
    "post",
    "postfix",
    "postmaster",
    "posts",
    "pr",
    "pr0n",
    "premium",
    "press",
    "price",
    "pricing",
    "principles",
    "print",
    "privacy",
    "privacy-policy",
    "privacypolicy",
    "privacy_policy",
    "private",
    "prod",
    "product",
    "production",
    "products",
    "profile",
    "profiles",
    "project",
    "projects",
    "promo",
    "promotions",
    "proxies",
    "proxy",
    "pub",
    "public",
    "purchase",
    "purpose",
    "put",
    "python",
    "queries",
    "query",
    "radio",
    "random",
    "ranking",
    "read",
    "reader",
    "readme",
    "recent",
    "recruit",
    "recruitment",
    "redirect",
    "register",
    "registration",
    "release",
    "remove",
    "replies",
    "report",
    "reports",
    "repositories",
    "repository",
    "req",
    "request",
    "requests",
    "research",
    "reset",
    "resolve",
    "resolver",
    "review",
    "rnail",
    "rnicrosoft",
    "roc",
    "root",
    "rss",
    "ruby",
    "rule",
    "sag",
    "sale",
    "sales",
    "sample",
    "samples",
    "sandbox",
    "save",
    "scholar",
    "school",
    "schools",
    "script",
    "scripts",
    "search",
    "secure",
    "security",
    "self",
    "seminars",
    "send",
    "server",
    "server-info",
    "server_info",
    "server-status",
    "server_status",
    "servers",
    "service",
    "services",
    "session",
    "sessions",
    "setting",
    "settings",
    "setup",
    "share",
    "shop",
    "shopping",
    "shortcut",
    "shortcuts",
    "show",
    "sign-in",
    "sign-up",
    "signin",
    "signout",
    "signup",
    "sign_in",
    "sign_up",
    "site",
    "sitemap",
    "sitemaps",
    "sitenews",
    "sites",
    "sketchup",
    "sky",
    "slash",
    "slashinvoice",
    "slut",
    "smartphone",
    "sms",
    "smtp",
    "soap",
    "software",
    "sorry",
    "source",
    "spec",
    "special",
    "spreadsheet",
    "spreadsheets",
    "sql",
    "src",
    "srntp",
    "ssh",
    "ssl",
    "ssladmin",
    "ssladministrator",
    "sslwebmaster",
    "ssytem",
    "staff",
    "stage",
    "staging",
    "start",
    "stat",
    "state",
    "static",
    "statistics",
    "stats",
    "status",
    "store",
    "stores",
    "stories",
    "style",
    "styleguide",
    "styles",
    "stylesheet",
    "stylesheets",
    "subdomain",
    "subscribe",
    "subscription",
    "subscriptions",
    "suggest",
    "suggestqueries",
    "support",
    "survey",
    "surveys",
    "surveytool",
    "svn",
    "swf",
    "syn",
    "sync",
    "syndicated",
    "sys",
    "sysadmin",
    "sysadministrator",
    "sysadmins",
    "system",
    "tablet",
    "tablets",
    "tag",
    "tags",
    "talk",
    "talkgadget",
    "task",
    "tasks",
    "team",
    "teams",
    "tech",
    "telnet",
    "term",
    "terms",
    "terms-of-service",
    "termsofservice",
    "terms_of_service",
    "test",
    "testing",
    "tests",
    "text",
    "theme",
    "themes",
    "thread",
    "threads",
    "ticket",
    "tickets",
    "tmp",
    "todo",
    "to-do",
    "to_do",
    "toml",
    "tool",
    "toolbar",
    "toolbars",
    "tools",
    "top",
    "topic",
    "topics",
    "tos",
    "tour",
    "trac",
    "translate",
    "trace",
    "translation",
    "translations",
    "translator",
    "trends",
    "tutorial",
    "tux",
    "tv",
    "twitter",
    "txt",
    "ul",
    "undef",
    "unfollow",
    "unsubscribe",
    "update",
    "updates",
    "upgrade",
    "upgrades",
    "upi",
    "upload",
    "uploads",
    "url",
    "usage",
    "user",
    "username",
    "usernames",
    "users",
    "uuid",
    "validation",
    "validations",
    "ver",
    "version",
    "video",
    "videos",
    "video-stats",
    "visitor",
    "visitors",
    "voice",
    "volunteer",
    "volunteers",
    "w",
    "watch",
    "wave",
    "weather",
    "web",
    "webdisk",
    "webhook",
    "webhooks",
    "webmail",
    "webmaster",
    "webmasters",
    "webrnail",
    "website",
    "websites",
    "welcome",
    "whm",
    "whois",
    "widget",
    "widgets",
    "wifi",
    "wiki",
    "wikis",
    "win",
    "windows",
    "word",
    "work",
    "works",
    "workshop",
    "wpad",
    "ww",
    "wws",
    "www",
    "wwws",
    "wwww",
    "xfn",
    "xhtml",
    "xhtrnl",
    "xml",
    "xmpp",
    "xpg",
    "xxx",
    "yaml",
    "year",
    "yml",
    "you",
    "yourdomain",
    "yourname",
    "yoursite",
    "yourusername",
]

FAMOUS_ACCOUNTS = [
    # reserving some large twitter accounts (top 100 by followers according to wikidata dump)
    "10ronaldinho",
    "3gerardpique",
    "adele",
    "akshaykumar",
    "aliaa08",
    "aliciakeys",
    "amitshah",
    "andresiniesta8",
    "anushkasharma",
    "arianagrande",
    "arrahman",
    "arvindkejriwal",
    "avrillavigne",
    "barackobama",
    "bbcbreaking",
    "bbcworld",
    "beingsalmankhan",
    "billgates",
    "britneyspears",
    "brunomars",
    "bts_bighit",
    "bts_twt",
    "championsleague",
    "chrisbrown",
    "cnnbrk",
    "coldplay",
    "conanobrien",
    "cristiano",
    "danieltosh",
    "davidguetta",
    "ddlovato",
    "deepikapadukone",
    "drake",
    "elisapie",
    "ellendegeneres",
    "elonmusk",
    "eminem",
    "emmawatson",
    "fcbarcelona",
    "foxnews",
    "harry_styles",
    "hillaryclinton",
    "iamsrk",
    "ihrithik",
    "imvkohli",
    "instagram",
    "jimmyfallon",
    "jlo",
    "joebiden",
    "jtimberlake",
    "justinbieber",
    "kaka",
    "kanyewest",
    "katyperry",
    "kendalljenner",
    "kevinhart4real",
    "khloekardashian",
    "kimkardashian",
    "kingjames",
    "kourtneykardash",
    "kyliejenner",
    "ladygaga",
    "liampayne",
    "liltunechi",
    "manutd",
    "mariahcarey",
    "mileycyrus",
    "mohamadalarefe",
    "narendramodi",
    "nasa",
    "nba",
    "neymarjr",
    "nfl",
    "niallofficial",
    "nickiminaj",
    "npr",
    "nytimes",
    "onedirection",
    "oprah",
    "pink",
    "pitbull",
    "playstation",
    "pmoindia",
    "premierleague",
    "priyankachopra",
    "realdonaldtrump",
    "ricky_martin",
    "rihanna",
    "sachin_rt",
    "selenagomez",
    "shakira",
    "shawnmendes",
    "sportscenter",
    "srbachchan",
    "subhisharma100",
    "taylorswift13",
    "theeconomist",
    "twitter",
    "virendersehwag",
    "whitehouse45",
    "wizkhalifa",
    "youtube",
    "zaynmalik",
    # some top instagram (https://en.wikipedia.org/wiki/List_of_most-followed_Instagram_accounts)
    "beyonce",
    "billieeilish",
    "leomessi",
    "natgeo",
    "nike",
    "snoopdogg",
    "taylorswift",
    "therock",
    # ... and a couple more prominent accounts, subjectively
    "10downingstreet",
    "aoc",
    "carterjwm",
    "dril",
    "gretathunberg",
    "kamalaharris",
    "kremlinrussia_e",
    "potus",
    "rondesantisfl",
    "ukraine",
    "washingtonpost",
    "yousuck2020",
    "zelenskyyua",
]


RESERVED_SUBDOMAINS = set(ATP_SPECIFIC + COMMONLY_RESERVED + FAMOUS_ACCOUNTS)



from pathlib import Path
import re
import textwrap
from typing import List
from functools import lru_cache, wraps


def to_lex_uri(string: str, base_uri: str = None) -> str:
    if string.startswith("lex:"):
        return string

    if string.startswith("#"):
        if not base_uri:
            raise ValueError(f"Unable to resolve URI without anchor: {string}")
        return f"{base_uri}{string}"

    return f"lex:{string}"


def memoized_property(func):
    """
    A decorator for creating memoized properties on Pydantic models.
    """
    @property
    @wraps(func)
    def wrapped(self):
        # Store the cached values in the model's __dict__
        if func.__name__ not in self.__dict__:
            self.__dict__[func.__name__] = func(self)
        return self.__dict__[func.__name__]

    return wrapped
    



# ==============================================================================

def ensure_path_and_init_files(path: Path):
    path_obj = Path(path)
    path_obj.mkdir(parents=True, exist_ok=True)
    init_file = path_obj / "__init__.py"
    init_file.touch()


def snake_to_camel(name: str) -> str:
    return "".join(x.capitalize() or "_" for x in name.split("_"))


def camel_to_snake(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def build_python_module_file_structure(output_dir: Path, id_str: str) -> Path:
    id_parts = id_str.split(".")
    package_names = id_parts[:-1]
    file_name = camel_to_snake(id_parts[-1]) + ".py"

    base_path = output_dir
    for package_name in package_names:
        base_path = base_path / package_name
        ensure_path_and_init_files(base_path)

    return base_path / file_name

def import_types_str(package_path: str, type_names: List[str], max_line=79) -> str:
    """
    Builds a string that imports the given type names from the given package path.
    Reformats the string to be within the given max_line length.
    """
    import_str = f'from {package_path} import {", ".join(type_names)}'
    if len(import_str) <= max_line:
        return import_str

    import_str = f"from {package_path} import (\n"
    import_str += "    " + textwrap.fill(
        ", ".join(type_names), max_line - 4, subsequent_indent="    "
    )
    import_str += "\n)"
    return import_str


def camel_to_class_name(s: str) -> str:
    return s[0].upper() + s[1:]


import re
from psychonaut.nsid.nsid import ensure_valid_nsid, InvalidNsidError, ensure_valid_nsid_regex
from psychonaut.identifier.handle import ensure_valid_handle, InvalidHandleError, ensure_valid_handle_regex
from psychonaut.identifier.did import ensure_valid_did, InvalidDidError, ensure_valid_did_regex

def ensure_valid_at_uri(uri: str):
    # JSON pointer is pretty different from rest of URI, so split that out first
    uri_parts = uri.split('#')
    if len(uri_parts) > 2:
        raise ValueError('ATURI can have at most one "#", separating fragment out')
    fragment_part = uri_parts[1] if len(uri_parts) == 2 else None
    uri = uri_parts[0]

    # check that all chars are boring ASCII
    if not re.match(r'^[a-zA-Z0-9._~:@!$&\')(*+,;=%/-]*$', uri):
        raise ValueError('Disallowed characters in ATURI (ASCII)')

    parts = uri.split('/')
    if len(parts) >= 3 and (parts[0] != 'at:' or len(parts[1]) != 0):
        raise ValueError('ATURI must start with "at://"')
    if len(parts) < 3:
        raise ValueError('ATURI requires at least method and authority sections')

    try:
        ensure_valid_handle(parts[2])
    except InvalidHandleError:
        try:
            ensure_valid_did(parts[2])
        except InvalidDidError:
            raise ValueError('ATURI authority must be a valid handle or DID')

    if len(parts) >= 4:
        if len(parts[3]) == 0:
            raise ValueError(
                'ATURI cannot have a slash after authority without a path segment'
            )
        try:
            ensure_valid_nsid(parts[3])
        except InvalidNsidError:
            raise ValueError(
                'ATURI requires first path segment (if supplied) to be valid NSID'
            )

    if len(parts) >= 5:
        if len(parts[4]) == 0:
            raise ValueError(
                'ATURI cannot have a slash after collection, unless record key is provided'
            )
        # would validate rkey here, but there are basically no constraints!

    if len(parts) >= 6:
        raise ValueError(
            'ATURI path can have at most two parts, and no trailing slash'
        )

    if len(uri_parts) >= 2 and fragment_part is None:
        raise ValueError('ATURI fragment must be non-empty and start with slash')

    if fragment_part is not None:
        if len(fragment_part) == 0 or fragment_part[0] != '/':
            raise ValueError('ATURI fragment must be non-empty and start with slash')
        # NOTE: enforcing *some* checks here for sanity. Eg, at least no whitespace
        if not re.match(r'^\/[a-zA-Z0-9._~:@!$&\')(*+,;=%[\]/-]*$', fragment_part):
            raise ValueError('Disallowed characters in ATURI fragment (ASCII)')

    if len(uri) > 8 * 1024:
        raise ValueError('ATURI is far too long')


_aturi_re = r"""^at:\/\/(?P<authority>[a-zA-Z0-9._:%-]+)(\/(?P<collection>[a-zA-Z0-9-.]+)(\/(?P<rkey>[a-zA-Z0-9._~:@!$&%')(*+,;=-]+))?)?(#(?P<fragment>\/[a-zA-Z0-9._~:@!$&%')(*+,;=\-[\]/\\]*))?$"""



def ensure_valid_at_uri_regex(uri: str):
    # simple regex to enforce most constraints via just regex and length.
    # hand wrote this regex based on above constraints. whew!
    aturi_regex = re.compile(_aturi_re)
    rm = aturi_regex.match(uri)
    if not rm or not rm.groupdict():
        raise ValueError("ATURI didn't validate via regex")
    groups = rm.groupdict()

    try:
        ensure_valid_handle_regex(groups['authority'])
    except ValueError:
        try:
            ensure_valid_did_regex(groups['authority'])
        except ValueError:
            raise ValueError('ATURI authority must be a valid handle or DID')

    if groups['collection']:
        try:
            ensure_valid_nsid_regex(groups['collection'])
        except ValueError:
            raise ValueError('ATURI collection path segment must be a valid NSID')

    if len(uri) > 8 * 1024:
        raise ValueError('ATURI is far too long')



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_cid, validate_did
from psychonaut.api.session import Session


class Repo(BaseModel):
    """
    [none provided by spec]
    """

    did: str = Field(..., pre=True, validator=validate_did)
    head: str = Field(..., pre=True, validator=validate_cid)


class ListReposResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    repos: Any


class ListReposReq(BaseModel):
    """
    List dids and root cids of hosted repos
    """

    limit: Optional[int] = Field(default=500, ge=1, le=1000)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.listRepos"

    async def do_xrpc(self, sess: Session) -> ListReposResp:
        resp = await sess.query(self)
        return ListReposResp(**resp)



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileView
from psychonaut.api.lexicons.com.atproto.label.defs import Label
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid, validate_datetime
from psychonaut.api.session import Session


class Notification(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    author: ProfileView
    reason: str = Field(
        ...,
        description="Expected values are 'like', 'repost', 'follow', 'mention', 'reply', and 'quote'.",
        known_values=["like", "repost", "follow", "mention", "reply", "quote"],
    )
    reasonSubject: Optional[str] = Field(
        default=None, pre=True, validator=validate_at_uri
    )
    record: Any
    isRead: bool = Field(...)
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    labels: Optional[Any] = None


class ListNotificationsResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    notifications: Any


class ListNotificationsReq(BaseModel):
    """
    [none provided by spec]
    """

    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)
    seenAt: Optional[str] = Field(default=None, pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.notification.listNotifications"

    async def do_xrpc(self, sess: Session) -> ListNotificationsResp:
        resp = await sess.query(self)
        return ListNotificationsResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_handle


class RefreshSessionResp(BaseModel):
    accessJwt: str = Field(...)
    refreshJwt: str = Field(...)
    handle: str = Field(..., pre=True, validator=validate_handle)
    did: str = Field(..., pre=True, validator=validate_did)


class RefreshSessionReq(BaseModel):
    """
    Refresh an authentication session.
    """

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.refreshSession"

    async def do_xrpc(self, sess: Session) -> RefreshSessionResp:
        resp = await sess.procedure(self)
        return RefreshSessionResp(**resp)



from typing import Any, Callable, Iterable

from pydantic import BaseModel
from carbox.car import read_car
from psychonaut.firehose.serde import read_event_pair, read_enriched_event


ERROR_CALLBACK_TYPE = Callable[[int, bytes], None]


def noop_error_callback(i: int, e: Exception, msg: bytes) -> None:
    pass


class ErrorCollector:
    def __init__(self):
        self.errors = []

    def __call__(self, i: int, e: Exception, msg: bytes) -> None:
        self.errors.append((i, e, msg))

    def __len__(self):
        return len(self.errors)


def iter_events(
    messages: Iterable[bytes], err_callback: ERROR_CALLBACK_TYPE = noop_error_callback
) -> Iterable[Any]:
    for i, msg in enumerate(messages):
        header, event = read_event_pair(msg)
        try:
            roots, blocks = read_car(event["blocks"])
            yield header, event, roots, blocks
        except KeyError as e:
            err_callback(i, e, msg)


def iter_enriched_events(
    messages: Iterable[bytes], err_callback: ERROR_CALLBACK_TYPE = noop_error_callback
) -> Iterable[BaseModel]:
    for i, msg in enumerate(messages):
        try:
            yield read_enriched_event(msg)
        except Exception as e:
            err_callback(i, e, msg)




from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid, validate_datetime
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileViewBasic
from psychonaut.api.lexicons.app.bsky.embed.images import View
from psychonaut.api.lexicons.app.bsky.embed.record_with_media import View
from psychonaut.api.lexicons.com.atproto.label.defs import Label
from psychonaut.api.lexicons.app.bsky.embed.record import View
from psychonaut.api.lexicons.app.bsky.embed.external import View


class NotFoundPost(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    notFound: bool = Field(...)


class BlockedPost(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    blocked: bool = Field(...)


class ViewerState(BaseModel):
    """
    [none provided by spec]
    """

    repost: Optional[str] = Field(default=None, pre=True, validator=validate_at_uri)
    like: Optional[str] = Field(default=None, pre=True, validator=validate_at_uri)


class ReasonRepost(BaseModel):
    """
    [none provided by spec]
    """

    by: ProfileViewBasic
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)


class PostView(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    author: ProfileViewBasic
    record: Any
    embed: Optional[Any] = None
    replyCount: Optional[int] = Field(default=None)
    repostCount: Optional[int] = Field(default=None)
    likeCount: Optional[int] = Field(default=None)
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    viewer: Optional[ViewerState] = None
    labels: Optional[Any] = None


class ReplyRef(BaseModel):
    """
    [none provided by spec]
    """

    root: PostView
    parent: PostView


class ThreadViewPost(BaseModel):
    """
    [none provided by spec]
    """

    post: PostView
    parent: Optional[Any] = None
    replies: Optional[List[Any]] = Field(default=None)


class FeedViewPost(BaseModel):
    """
    [none provided by spec]
    """

    post: PostView
    reply: Optional[ReplyRef] = None
    reason: Optional[Any] = None



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileView
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid, validate_datetime
from psychonaut.api.session import Session


class Like(BaseModel):
    """
    [none provided by spec]
    """

    indexedAt: str = Field(..., pre=True, validator=validate_datetime)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)
    actor: ProfileView


class GetLikesResp(BaseModel):
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)
    cursor: Optional[str] = Field(default=None)
    likes: Any


class GetLikesReq(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)
    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.getLikes"

    async def do_xrpc(self, sess: Session) -> GetLikesResp:
        resp = await sess.query(self)
        return GetLikesResp(**resp)



from psychonaut.api.session import Session


async def collect_cursored(
    sess: Session,
    req,
    resp_type,
    collection_k: str,
    cursor_key="cursor",
):
    cursor, last_cursor = None, None
    while True:
        req = req.copy(update={cursor_key: cursor})
        resp = await req.do_xrpc(sess)
        assert isinstance(resp, resp_type)
        cursor = resp.cursor

        new_items = getattr(resp, collection_k)
        for item in new_items:
            yield item

        if not cursor or cursor == last_cursor or not new_items:
            break

        last_cursor = cursor


from graphlib import TopologicalSorter
from pathlib import Path
from typing import Dict, List, Optional, Set, Union
from pydantic import BaseModel, Field
from psychonaut.lexicon.types import (
    LexArray,
    LexRef,
    LexRefUnion,
    LexString,
    LexiconDoc,
)
from psychonaut.lexicon.util import (
    camel_to_snake,
    memoized_property,
    camel_to_class_name,
)


class LexiconDef(BaseModel):
    node_id: str
    deps: Set[str] = Field(default_factory=set)
    input_path: Path
    defn: BaseModel = Field(repr=False)

    generated_src: Optional[str] = Field(default=None, repr=False)

    @memoized_property
    def leaf_id(self) -> str:
        parts = self.node_id.split("#")
        if len(parts) == 2:
            return parts[1]

        parts = self.node_id.split(".")
        if len(parts) > 1:
            return parts[-1]

        raise ValueError(f"Unable to determine leaf id for {self.node_id}")

    @memoized_property
    def leaf_id_snake(self) -> str:
        return camel_to_snake(self.leaf_id)

    @memoized_property
    def leaf_id_as_class_name(self) -> str:
        return camel_to_class_name(self.leaf_id)


def create_plan(input_dir: Path):
    for node in compile_toposorted_lexicon_defns(input_dir):
        print(node.node_id, node.defn.type)


def compile_toposorted_lexicon_defns(input_dir: Path) -> List[LexiconDef]:
    def_id_to_lexicon_def = {}
    graph: Dict[str, Set[str]] = {}

    # Collect all the defs in the first pass.
    for input_path in input_dir.glob("**/*.json"):
        # Parse the document once.
        doc = LexiconDoc.parse_file(input_path)

        parent_xrpc_id = doc.id

        # Anything at this level has a defs section.
        for k, defn in doc.defs.items():
            # maybe NSID? but with fragment?
            def_id = f"{doc.id}#{k}" if k != "main" else doc.id

            graph[def_id] = _extract_deps_from_base_model(parent_xrpc_id, def_id, defn)

            def_id_to_lexicon_def[def_id] = LexiconDef(
                node_id=def_id,
                deps=graph[def_id].copy(),
                input_path=input_path,
                defn=defn,
            )

    # Do topo sort to check for cycles first.
    nodes = list(TopologicalSorter(graph).static_order())

    # Identify missing. If there are any, we can't continue.
    missing = set(n for n in nodes if n not in def_id_to_lexicon_def)
    if missing:
        print("MISSING DEFINITIONS FOR:")
        for k in missing:
            print(f"\t{k}")
        return []

    return [def_id_to_lexicon_def[k] for k in nodes]


def _extract_deps_from_base_model(
    parent_xrpc_id: str, def_id: str, defn: BaseModel
) -> Set[str]:
    deps = set()

    for field in defn.__fields__:
        _extract_deps(parent_xrpc_id, getattr(defn, field, None), deps)

    if def_id in deps:
        deps.remove(def_id)  # cycle

    if def_id == "app.bsky.embed.record#viewRecord":
        print("MANGLING app.bsky.embed.record#viewRecord cycle")
        return set()  # TODO: cycle break kludge

    return deps


def _extract_deps(
    parent_xrpc_id: str, defn, deps: Optional[Set[str]] = None
) -> Set[str]:
    if deps is None:
        deps = set()

    if defn is None:
        return deps

    if isinstance(defn, list) and len(defn) > 0 and isinstance(defn[0], str):
        # Kludge for knownValues
        for item in defn:
            if "#" in item:
                deps.add(_fully_qualified_ref(parent_xrpc_id, item))

    if not isinstance(defn, dict):
        # TODO: ugly as fuck
        if isinstance(defn, BaseModel):
            for field in defn.__fields__:
                _extract_deps(parent_xrpc_id, getattr(defn, field, None), deps)
            return
        else:
            return deps

    for k, v in defn.items():
        # If the value is a dict, recurse
        if isinstance(v, dict):
            _extract_deps(parent_xrpc_id, v, deps)
        # If the value is a list, recurse
        elif isinstance(v, list):
            for item in v:
                _extract_deps(parent_xrpc_id, item, deps)
        elif isinstance(v, (LexRef, LexRefUnion)):
            _extract_ref_deps_from(deps, parent_xrpc_id, v)
        elif isinstance(v, LexArray):
            _extract_ref_deps_from(deps, parent_xrpc_id, v.items)
        elif isinstance(v, LexString):
            _extract_ref_deps_from_lex_string(deps, parent_xrpc_id, v)
        elif isinstance(v, str):
            raise "A"
            if v.startswith("#"):
                deps.add(_fully_qualified_ref(parent_xrpc_id, v))

    return deps


def _extract_ref_deps_from(
    deps: Set[str], parent_xrpc_id, v: Union[LexRefUnion, LexRef]
) -> Set[str]:
    if isinstance(v, LexRef):
        deps.add(_fully_qualified_ref(parent_xrpc_id, v.ref))
    elif isinstance(v, LexRefUnion):
        deps.update(_fully_qualified_ref(parent_xrpc_id, ref) for ref in v.refs)


def _extract_ref_deps_from_lex_string(
    deps: Set[str], parent_xrpc_id: str, v: LexString
):
    if v.known_values:
        known_values = {
            _fully_qualified_ref(parent_xrpc_id, item)
            for item in v.known_values
            if "#" in item
        }
        if known_values:
            deps.update(known_values)


def _fully_qualified_ref(parent_xrpc_id: str, ref: str) -> str:
    if ref.startswith("#"):
        ref = f"{parent_xrpc_id}{ref}"
    return ref



from typing import Any, Optional
from psychonaut.api.lexicons.com.atproto.moderation.defs import ReasonType
from psychonaut.api.lexicons.com.atproto.admin.defs import RepoRef
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_datetime


class CreateReportResp(BaseModel):
    id: int = Field(...)
    reasonType: ReasonType
    reason: Optional[str] = Field(default=None)
    subject: Any
    reportedBy: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)


class CreateReportReq(BaseModel):
    """
    Report a repo or a record.
    """

    reasonType: ReasonType
    reason: Optional[str] = Field(default=None)
    subject: Any

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.moderation.createReport"

    async def do_xrpc(self, sess: Session) -> CreateReportResp:
        resp = await sess.procedure(self)
        return CreateReportResp(**resp)



from typing import Any, Optional, List
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did


class ResolveModerationReportsReq(BaseModel):
    """
    Resolve moderation reports by an action.
    """

    actionId: int = Field(...)
    reportIds: List[int] = Field(...)
    createdBy: str = Field(..., pre=True, validator=validate_did)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.resolveModerationReports"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.api.lexicons.com.atproto.label.defs import Label


class Info(BaseModel):
    """
    [none provided by spec]
    """

    name: str = Field(..., known_values=["OutdatedCursor"])
    message: Optional[str] = Field(default=None)


class Labels(BaseModel):
    """
    [none provided by spec]
    """

    seq: int = Field(...)
    labels: Any



from enum import Enum
from typing import Optional, List, Union
from multiformats import CID
from pydantic import BaseModel, Field

from psychonaut.lexicon.formats import validate_at_uri, validate_cid


# Event stream messages =====================================


class WriteOpAction(str, Enum):
    Create = "Create"
    Update = "Update"


# TODO: CID and URL ddmake sense as types
class IndexRecord(BaseModel):
    type: str = "index_record"
    action: str = Field(..., pre=True, validator=WriteOpAction.__contains__)
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: CID = Field(..., pre=True, validator=validate_cid)
    obj: object
    timestamp: str

    class Config:
        json_encoders = { CID: str }
        arbitrary_types_allowed = True



class DeleteRecord(BaseModel):
    type: str = "delete_record"
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cascading: bool


class DeleteRepo(BaseModel):
    type: str = "delete_repo"
    did: str


# Sequencer events =========================================
# /pds/src/sequencer/events.ts


class ActionType(str, Enum):
    create = "create"
    update = "update"
    delete = "delete"


class CommitEvtOp(BaseModel):
    action: ActionType
    path: str
    cid: Optional[CID] = Field(default=None, pre=True, validator=validate_cid)

    class Config:
        json_encoders = { CID: str }
        arbitrary_types_allowed = True


class CommitEvt(BaseModel):
    repo: str
    ops: List[CommitEvtOp]
    commit: CID = Field(..., pre=True, validator=validate_cid)
    prev: Optional[CID] = Field(default=None, pre=True, validator=validate_cid)
    blocks: bytes
    blobs: List[str] = Field(default_factory=list, pre=True, validator=validate_cid)
    rebase: bool
    too_big: bool = Field(..., alias="tooBig")

    class Config:
        json_encoders = { CID: str }
        arbitrary_types_allowed = True


class HandleEvt(BaseModel):
    did: str
    handle: str


class TypedCommitEvt(BaseModel):
    type: str = Field("commit", const=True)
    seq: int
    time: str
    evt: CommitEvt


class TypedHandleEvt(BaseModel):
    type: str = Field("handle", const=True)
    seq: int
    time: str
    evt: HandleEvt


SeqEvt = Union[TypedCommitEvt, TypedHandleEvt]



from typing import Any, Optional, List
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_cid, validate_did, validate_array


class GetBlocksReq(BaseModel):
    """
    Gets blocks from a given repo.
    """

    did: str = Field(
        ..., description="The DID of the repo.", pre=True, validator=validate_did
    )
    cids: List[str] = Field(..., pre=True, validator=validate_array(validate_cid))

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.getBlocks"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_uri


class External(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_uri)
    title: str = Field(...)
    description: str = Field(...)
    thumb: Optional[Any] = None


class ViewExternal(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_uri)
    title: str = Field(...)
    description: str = Field(...)
    thumb: Optional[str] = Field(default=None)


class External(BaseModel):
    """
    [none provided by spec]
    """

    external: External


class View(BaseModel):
    """
    [none provided by spec]
    """

    external: ViewExternal



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_cid, validate_did


class GetHeadResp(BaseModel):
    root: str = Field(..., pre=True, validator=validate_cid)


class GetHeadReq(BaseModel):
    """
    Gets the current HEAD CID of a repo.
    """

    did: str = Field(
        ..., description="The DID of the repo.", pre=True, validator=validate_did
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.getHead"

    async def do_xrpc(self, sess: Session) -> GetHeadResp:
        resp = await sess.query(self)
        return GetHeadResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_datetime


class CreateAppPasswordReq(BaseModel):
    """
    Create an app-specific password.
    """

    name: str = Field(...)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.createAppPassword"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)


class AppPassword(BaseModel):
    """
    [none provided by spec]
    """

    name: str = Field(...)
    password: str = Field(...)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_handle, validate_did


class UpdateAccountHandleReq(BaseModel):
    """
    Administrative action to update an account's handle
    """

    did: str = Field(..., pre=True, validator=validate_did)
    handle: str = Field(..., pre=True, validator=validate_handle)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.updateAccountHandle"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



import pytest
from .resolve import resolve_dns, NoHandleRecordError


@pytest.mark.asyncio
async def test_resolve_dns():
    handle = "generativist.xyz"
    expected_did = "did:plc:o32okshy54r5h2vlrjpz3aln"

    did = await resolve_dns(handle)
    assert did == expected_did


@pytest.mark.asyncio
async def test_resolve_dns_no_handle_record_error():
    handle = "nope.generativist.xyz"

    with pytest.raises(NoHandleRecordError):
        await resolve_dns(handle)


import json
import click
from psychonaut.api.lexicons.app.bsky.graph.get_followers import (
    GetFollowersReq, GetFollowersResp
)
from psychonaut.api.lexicons.app.bsky.graph.get_follows import (
    GetFollowsReq, GetFollowsResp,
)
from psychonaut.cli.group import cli
from psychonaut.cli.util import as_async, clean_handle
from psychonaut.client import get_simple_client_session
from psychonaut.client.cursors import collect_cursored


@cli.command()
@click.argument("actor")
@as_async
async def get_followers(actor: str):
    actor = clean_handle(actor)

    async with get_simple_client_session() as sess:
        gen = collect_cursored(
            sess,
            GetFollowersReq(actor=actor, limit=100),
            GetFollowersResp,
            "followers"
        )

        async for follower in gen:
            print(json.dumps(follower))


@cli.command()
@click.argument("actor")
@as_async
async def get_follows(actor: str):
    actor = clean_handle(actor)

    async with get_simple_client_session() as sess:
        gen = collect_cursored(
            sess,
            GetFollowsReq(actor=actor, limit=100),
            GetFollowsResp,
            "follows"
        )

        async for follows in gen:
            print(json.dumps(follows))


from typing import Any, Optional, List
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class DisableInviteCodesReq(BaseModel):
    """
    Disable some set of codes and/or all codes associated with a set of
    users
    """

    codes: Optional[List[str]] = Field(default=None)
    accounts: Optional[List[str]] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.disableInviteCodes"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.api.session import Session
from psychonaut.lexicon.formats import validate_did, validate_array


class AccountCodes(BaseModel):
    """
    [none provided by spec]
    """

    account: str = Field(...)
    codes: List[str] = Field(...)


class CreateInviteCodesResp(BaseModel):
    codes: Any


class CreateInviteCodesReq(BaseModel):
    """
    Create an invite code.
    """

    codeCount: int = Field(default=1)
    useCount: int = Field(...)
    forAccounts: Optional[List[str]] = Field(
        default=None, pre=True, validator=validate_array(validate_did)
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.createInviteCodes"

    async def do_xrpc(self, sess: Session) -> CreateInviteCodesResp:
        resp = await sess.procedure(self)
        return CreateInviteCodesResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class RequestCrawlReq(BaseModel):
    """
    Request a service to persistently crawl hosted repos.
    """

    hostname: str = Field(
        ..., description="Hostname of the service that is requesting to be crawled."
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.sync.requestCrawl"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



"""
More or less a GPT-4 translation of the NSID code from:
https://github.com/bluesky-social/atproto/blob/656be937a52688601c088f5aa332137dfaf019cf/packages/nsid/src/index.ts
"""
import re
from pydantic import BaseModel, validator
from typing import List

class NSID(BaseModel):
    """
    Namespaced IDs
    """
    segments: List[str]

    @classmethod
    def parse(cls, nsid: str):
        ensure_valid_nsid(nsid)
        segments = nsid.split('.')
        return cls(segments=segments)

    @classmethod
    def create(cls, authority: str, name: str):
        segments = authority.split('.')[::-1] + [name]
        nsid = '.'.join(segments)
        return cls.parse(nsid)

    @classmethod
    def is_valid(cls, nsid: str) -> bool:
        try:
            cls.parse(nsid)
            return True
        except InvalidNsidError:
            return False

    @validator("segments", pre=True, always=True)
    def validate_segments(cls, nsid):
        if isinstance(nsid, str):
            ensure_valid_nsid(nsid)
            segments = nsid.split('.')
            return segments
        return nsid

    @property
    def authority(self):
        return '.'.join(self.segments[:-1][::-1])

    @property
    def name(self):
        return self.segments[-1]

    def __str__(self):
        return '.'.join(self.segments)



class InvalidNsidError(Exception):
    pass


def ensure_valid_nsid(nsid: str) -> None:
    split = nsid.split(".")
    to_check = ".".join(split[:-1]) if split[-1] == "*" else ".".join(split)

    if not re.match(r"^[a-zA-Z0-9.-]*$", to_check):
        raise InvalidNsidError(
            "Disallowed characters in NSID (ASCII letters, digits, dashes, periods only)"
        )

    if len(to_check) > 253 + 1 + 128:
        raise InvalidNsidError("NSID is too long (382 chars max)")

    labels = to_check.split(".")
    if len(split) < 3:
        raise InvalidNsidError("NSID needs at least three parts")

    for i, l in enumerate(labels):
        if len(l) < 1:
            raise InvalidNsidError("NSID parts can not be empty")

        if len(l) > 63 and i + 1 < len(labels):
            raise InvalidNsidError("NSID domain part too long (max 63 chars)")

        if len(l) > 128 and i + 1 == len(labels):
            raise InvalidNsidError("NSID name part too long (max 127 chars)")

        if l.endswith("-"):
            raise InvalidNsidError("NSID parts can not end with hyphen")

        if not re.match(r"^[a-zA-Z]", l):
            raise InvalidNsidError("NSID parts must start with ASCII letter")


def ensure_valid_nsid_regex(nsid: str) -> None:
    if not re.match(
        (
            r"^[a-zA-Z]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?"
            r"(\.[a-zA-Z]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)+"
            r"(\.[a-zA-Z]([a-zA-Z0-9-]{0,126}[a-zA-Z0-9])?)$"
        ),
        nsid,
    ):
        raise InvalidNsidError("NSID didn't validate via regex")

    if len(nsid) > 253 + 1 + 128:
        raise InvalidNsidError("NSID is too long (382 chars max)")



import pytest
from .did import ensure_valid_did, ensure_valid_did_regex, InvalidDidError


def test_did_permissive_validation():
    expect_valid("did:method:val")
    expect_valid("did:method:VAL")
    expect_valid("did:method:val123")
    expect_valid("did:method:123")
    expect_valid("did:method:val-two")
    expect_valid("did:method:val_two")
    expect_valid("did:method:val.two")
    expect_valid("did:method:val:two")
    expect_valid("did:method:val%BB")

    expect_invalid("did")
    expect_invalid("didmethodval")
    expect_invalid("method:did:val")
    expect_invalid("did:method:")
    expect_invalid("didmethod:val")
    expect_invalid("did:methodval")
    expect_invalid(":did:method:val")
    expect_invalid("did.method.val")
    expect_invalid("did:method:val:")
    expect_invalid("did:method:val%")
    expect_invalid("DID:method:val")
    expect_invalid("did:METHOD:val")
    expect_invalid("did:m123:val")

    expect_valid("did:method:" + "v" * 240)
    expect_invalid("did:method:" + "v" * 8500)

    expect_valid("did:m:v")
    expect_valid("did:method::::val")
    expect_valid("did:method:-")
    expect_valid("did:method:-:_:.:%ab")
    expect_valid("did:method:.")
    expect_valid("did:method:_")
    expect_valid("did:method::.")

    expect_invalid("did:method:val/two")
    expect_invalid("did:method:val?two")
    expect_invalid("did:method:val#two")
    expect_invalid("did:method:val%")

    expect_valid("did:onion:2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid")


def test_allows_some_real_did_values():
    expect_valid("did:example:123456789abcdefghi")
    expect_valid("did:plc:7iza6de2dwap2sbkpav7c6c6")
    expect_valid("did:web:example.com")
    expect_valid("did:key:zQ3shZc2QzApp2oymGvQbzP8eKheVshBHbU4ZYjeXqwSKEn6N")
    expect_valid("did:ethr:0xb9c5714089478a327f09197987f16f9e5d936e8a")


def expect_valid(h: str):
    ensure_valid_did(h)
    ensure_valid_did_regex(h)


def expect_invalid(h: str):
    with pytest.raises(InvalidDidError):
        ensure_valid_did(h)
    with pytest.raises(InvalidDidError):
        ensure_valid_did_regex(h)



from typing import Any, Optional, List
from psychonaut.api.lexicons.com.atproto.label.defs import Label
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_array


class QueryLabelsResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    labels: Any


class QueryLabelsReq(BaseModel):
    """
    Find labels relevant to the provided URI patterns.
    """

    uriPatterns: List[str] = Field(
        ...,
        description="List of AT URI patterns to match (boolean 'OR'). Each may be a prefix (ending with '*'; will match inclusive of the string leading to '*'), or a full URI",
    )
    sources: Optional[List[str]] = Field(
        default=None,
        description="Optional list of label sources (DIDs) to filter on",
        pre=True,
        validator=validate_array(validate_did),
    )
    limit: Optional[int] = Field(default=50, ge=1, le=250)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.label.queryLabels"

    async def do_xrpc(self, sess: Session) -> QueryLabelsResp:
        resp = await sess.query(self)
        return QueryLabelsResp(**resp)



from typing import Tuple
import click
from psychonaut.cli.group import cli
from psychonaut.cli.util import as_async
from psychonaut.client import get_simple_client_session
from psychonaut.api.lexicons.app.bsky.feed.post import Post

# from psychonaut.api.lexicons.com.atproto.repo.upload_blob import (
#     UploadBlobReq,
#     upload_blob as upload_blob_f,
#     UploadBlobResp,
# )
from datetime import datetime
from pathlib import Path


@cli.command()
@click.argument("text")
@click.option(
    "--image",
    "-i",
    multiple=True,
    help="image(s) to poast",
    type=click.Path(exists=True),
)
@as_async
async def poast(text: str, image: Tuple[str]):
    if image:
        # Verify each one exists
        for path in image:
            if not Path(path).exists():
                print(f"Image {path} does not exist")
                return

    async with get_simple_client_session() as sess:
        req = Post(text=text, createdAt=datetime.utcnow().isoformat())

        if image:
            images = []
            for img in image:
                content_type = "image/jpeg"  # TODO: get content type robustly
                if img.endswith(".png"):
                    content_type = "image/png"
                elif img.endswith(".gif"):
                    content_type = "image/gif"
                elif img.endswith(".webp"):
                    content_type = "image/webp"

                with open(img, "rb") as f:
                    img_bytes = f.read()

                resp = (await sess._post_blob_kludge(img_bytes, content_type))["blob"]
                images.append(resp)

            embed = {
                "$type": "app.bsky.embed.images",
                "images": [
                    {"alt": "", "image": img} for img in images  # TODO: get alt text
                ],
            }
            req = req.copy(update={"embed": embed})

        resp = await sess.record(req)
        print(resp)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_handle


class CreateSessionResp(BaseModel):
    accessJwt: str = Field(...)
    refreshJwt: str = Field(...)
    handle: str = Field(..., pre=True, validator=validate_handle)
    did: str = Field(..., pre=True, validator=validate_did)
    email: Optional[str] = Field(default=None)


class CreateSessionReq(BaseModel):
    """
    Create an authentication session.
    """

    identifier: str = Field(
        ...,
        description="Handle or other identifier supported by the server for the authenticating user.",
    )
    password: str = Field(...)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.createSession"

    async def do_xrpc(self, sess: Session) -> CreateSessionResp:
        resp = await sess.procedure(self)
        return CreateSessionResp(**resp)



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_cid,
    validate_at_identifier,
    validate_nsid,
)
from psychonaut.api.session import Session


class Delete(BaseModel):
    """
    Delete an existing record.
    """

    collection: str = Field(..., pre=True, validator=validate_nsid)
    rkey: str = Field(...)


class Create(BaseModel):
    """
    Create a new record.
    """

    collection: str = Field(..., pre=True, validator=validate_nsid)
    rkey: Optional[str] = Field(default=None)
    value: Any


class Update(BaseModel):
    """
    Update an existing record.
    """

    collection: str = Field(..., pre=True, validator=validate_nsid)
    rkey: str = Field(...)
    value: Any


class ApplyWritesReq(BaseModel):
    """
    Apply a batch transaction of creates, updates, and deletes.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    validate_flag: Optional[bool] = Field(
        alias="validate", default=True, description="Validate the records?"
    )
    writes: List[Any] = Field(...)
    swapCommit: Optional[str] = Field(default=None, pre=True, validator=validate_cid)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.applyWrites"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_at_uri, validate_cid, validate_datetime
from psychonaut.api.lexicons.app.bsky.actor.defs import ProfileViewBasic
from psychonaut.api.lexicons.com.atproto.repo.strong_ref import StrongRef


class ViewBlocked(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)


class ViewNotFound(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)


class ViewRecord(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    author: ProfileViewBasic
    value: Any
    labels: Optional[Any] = None
    embeds: Optional[List[Any]] = Field(default=None)
    indexedAt: str = Field(..., pre=True, validator=validate_datetime)


class Record(BaseModel):
    """
    [none provided by spec]
    """

    record: StrongRef


class View(BaseModel):
    """
    [none provided by spec]
    """

    record: Any



from typing import Any, Optional
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_at_uri,
    validate_cid,
    validate_at_identifier,
    validate_nsid,
)
from psychonaut.api.session import Session


class Record(BaseModel):
    """
    [none provided by spec]
    """

    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: str = Field(..., pre=True, validator=validate_cid)
    value: Any


class ListRecordsResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    records: Any


class ListRecordsReq(BaseModel):
    """
    List a range of records in a collection.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    collection: str = Field(
        ...,
        description="The NSID of the record type.",
        pre=True,
        validator=validate_nsid,
    )
    limit: Optional[int] = Field(
        default=50, description="The number of records to return.", ge=1, le=100
    )
    cursor: Optional[str] = Field(default=None)
    rkeyStart: Optional[str] = Field(
        default=None,
        description="DEPRECATED: The lowest sort-ordered rkey to start from (exclusive)",
    )
    rkeyEnd: Optional[str] = Field(
        default=None,
        description="DEPRECATED: The highest sort-ordered rkey to stop at (exclusive)",
    )
    reverse: Optional[bool] = Field(
        default=None, description="Reverse the order of the returned records?"
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.listRecords"

    async def do_xrpc(self, sess: Session) -> ListRecordsResp:
        resp = await sess.query(self)
        return ListRecordsResp(**resp)



import json
from typing import Tuple
import click

from psychonaut.cli.util import clean_handle
from .group import cli
from .util import as_async, print_error_and_fail
from psychonaut.api.lexicons.app.bsky.actor.get_profiles import GetProfilesReq
from psychonaut.api.lexicons.app.bsky.actor.get_profile import GetProfileReq
from psychonaut.api.lexicons.com.atproto.identity.resolve_handle import ResolveHandleReq
from psychonaut.client import get_simple_client_session


@cli.command()
@click.argument("handle")
@as_async
async def resolve_handle(handle: str):
    handle = clean_handle(handle)

    if not handle:
        print_error_and_fail("No handle provided")

    async with get_simple_client_session() as sess:
        resp = await ResolveHandleReq(handle=handle).do_xrpc(sess)
        print(resp.json())


@cli.command()
@click.argument("actors", nargs=-1)
@as_async
async def get_profiles(actors: Tuple[str]):
    actors = [clean_handle(actor) for actor in actors if clean_handle(actor)]

    if not actors:
        print_error_and_fail("No actors provided")

    async with get_simple_client_session() as sess:
        resp = await GetProfilesReq(actors=actors).do_xrpc(sess)
        for profile in resp.profiles:
            print(json.dumps(profile))


@cli.command()
@click.argument("actor", nargs=1)
@as_async
async def get_profile(actor: str):
    actor = clean_handle(actor)
    if not actor:
        print_error_and_fail("No actor provided")

    async with get_simple_client_session() as sess:
        resp = await GetProfileReq(actor=actor).do_xrpc(sess)
        print(json.dumps(resp))



"""
See:

https://github.com/bluesky-social/atproto/blob/656be937a52688601c088f5aa332137dfaf019cf/packages/lexicon/src/types.ts
"""
from typing import Any, Dict, Union, Optional, List
from pydantic import BaseModel, Field, validator, root_validator
from pydantic import ValidationError as PydanticValidationError
from psychonaut.nsid import NSID


# Primitive types =============================================================

# TODO: forbid extra?

class _BaseLex(BaseModel):
    class Config:
        extra = "forbid"


class LexBoolean(_BaseLex):
    type: str = Field("boolean", const=True)
    description: Optional[str]
    default: Optional[bool]
    const: Optional[bool]


class LexInteger(_BaseLex):
    type: str = Field("integer", const=True)
    description: Optional[str]
    default: Optional[int]
    minimum: Optional[int]
    maximum: Optional[int]
    enum: Optional[List[int]]
    const: Optional[int]


class LexStringFormat(str):
    datetime = "datetime"
    uri = "uri"
    at_uri = "at-uri"
    did = "did"
    handle = "handle"
    at_identifier = "at-identifier"
    nsid = "nsid"
    cid = "cid"


class LexString(_BaseLex):
    type: str = Field("string", const=True)
    format: Optional[LexStringFormat]
    description: Optional[str]
    default: Optional[str]
    min_length: Optional[int] = Field(None, alias="minLength")
    max_length: Optional[int] = Field(None, alias="maxLength")
    min_graphemes: Optional[int] = Field(None, alias="minGraphemes")
    max_graphemes: Optional[int] = Field(None, alias="maxGraphemes")
    enum: Optional[List[str]]
    const: Optional[str]
    known_values: Optional[List[str]] = Field(None, alias="knownValues")


class LexUnknown(_BaseLex):
    type: str = Field("unknown", const=True)
    description: Optional[str]


LexPrimitive = Union[LexBoolean, LexInteger, LexString, LexUnknown]

# ip1d types ==================================================================


class LexBytes(_BaseLex):
    type: str = Field("bytes", const=True)
    description: Optional[str]
    maxLength: Optional[int]
    minLength: Optional[int]


class LexCidLink(_BaseLex):
    type: str = Field("cid-link", const=True)
    description: Optional[str]


LexIpldType = Union[LexBytes, LexCidLink]


# reference types =============================================================


class LexRef(_BaseLex):
    type: str = Field("ref", const=True)
    description: Optional[str]
    ref: str


class LexRefUnion(_BaseLex):
    type: str = Field("union", const=True)
    description: Optional[str]
    refs: List[str]
    closed: Optional[bool]


LexRefVariant = Union[LexRef, LexRefUnion]

# blob types ==================================================================


class LexBlob(_BaseLex):
    type: str = Field("blob", const=True)
    description: Optional[str]
    accept: Optional[List[str]]
    maxSize: Optional[int]


# complex types ===============================================================


class LexArray(_BaseLex):
    type: str = Field("array", const=True)
    description: Optional[str]
    items: Union[LexPrimitive, LexIpldType, LexBlob, LexRefVariant]
    min_length: Optional[int] = Field(None, alias="minLength")
    max_length: Optional[int] = Field(None, alias="maxLength")


class LexPrimitiveArray(LexArray):
    items: LexPrimitive


class LexToken(_BaseLex):
    type: str = Field("token", const=True)
    description: Optional[str]


class LexObject(_BaseLex):
    type: str = Field("object", const=True)
    description: Optional[str]
    required: Optional[List[str]]
    nullable: Optional[List[str]]
    properties: Optional[
        dict[str, Union[LexRefVariant, LexIpldType, LexArray, LexBlob, LexPrimitive]]
    ]


# xrpc types =================================================================


class LexXrpcParameters(_BaseLex):
    type: str = Field("params", const=True)
    description: Optional[str]
    required: Optional[List[str]]
    properties: dict[str, Union[LexPrimitive, LexPrimitiveArray]]


class LexXrpcBody(_BaseLex):
    description: Optional[str]
    encoding: str
    schema_kludge: Optional[Union[LexRefVariant, LexObject]] = Field(
        None, alias="schema"
    )


class LexXrpcSubscriptionMessage(_BaseLex):
    description: Optional[str]
    schema_kludge: Optional[Union[LexRefVariant, LexObject]] = Field(
        None, alias="schema"
    )



class LexXrpcError(_BaseLex):
    name: str
    description: Optional[str]


class LexXrpcQuery(_BaseLex):
    type: str = Field("query", const=True)
    description: Optional[str]
    parameters: Optional[LexXrpcParameters]
    output: Optional[LexXrpcBody]
    errors: Optional[List[LexXrpcError]]


class LexXrpcProcedure(_BaseLex):
    type: str = Field("procedure", const=True)
    description: Optional[str]
    parameters: Optional[LexXrpcParameters]
    input: Optional[LexXrpcBody]
    output: Optional[LexXrpcBody]
    errors: Optional[List[LexXrpcError]]


class LexXrpcSubscription(_BaseLex):
    type: str = Field("subscription", const=True)
    description: Optional[str]
    parameters: Optional[LexXrpcParameters]
    message: Optional[LexXrpcSubscriptionMessage]
    infos: Optional[List[LexXrpcError]]
    errors: Optional[List[LexXrpcError]]


# database types ==============================================================


class LexRecord(_BaseLex):
    type: str = Field("record", const=True)
    description: Optional[str]
    key: Optional[str]
    record: LexObject


# core types ==================================================================


LexUserType = Union[
    LexRecord,
    LexXrpcQuery,
    LexXrpcProcedure,
    LexXrpcSubscription,
    LexBlob,
    LexArray,
    LexToken,
    LexObject,
    LexBoolean,
    LexInteger,
    LexString,
    LexBytes,
    LexCidLink,
    LexUnknown,
]

_lex_user_type_to_type = {
    "record": LexRecord,
    "query": LexXrpcQuery,
    "procedure": LexXrpcProcedure,
    "subscription": LexXrpcSubscription,
    "blob": LexBlob,
    "array": LexArray,
    "token": LexToken,
    "object": LexObject,
    "boolean": LexBoolean,
    "integer": LexInteger,
    "string": LexString,
    "bytes": LexBytes,
    "cid-link": LexCidLink,
    "unknown": LexUnknown,
}


class LexiconDoc(_BaseLex):
    lexicon: int = Field(1, const=True)
    id: str
    revision: Optional[int]
    description: Optional[str]
    defs: dict[str, LexUserType]

    @root_validator
    def validate_main_definition(cls, values):
        defs = values.get("defs")
        if defs:
            for def_id, def_ in defs.items():
                if def_id != "main" and (
                    def_.type == "record"
                    or def_.type == "procedure"
                    or def_.type == "query"
                    or def_.type == "subscription"
                ):
                    raise ValidationError(
                        "Records, procedures, queries, and subscriptions must be the main definition."
                    )
        return values

    @validator("id")
    def validate_nsid(cls, value):
        if not NSID.is_valid(value):
            raise ValidationError("Must be a valid NSID")
        return value

    def __init__(self, **data):
        if "defs" in data:
            data["defs"] = {
                k: _lex_user_type_to_type[v["type"]].parse_obj(v)
                for k, v in data["defs"].items()
            }
        super().__init__(**data)


# helpers =====================================================================


def is_valid_lexicon_doc(value: Any) -> bool:
    try:
        LexiconDoc.parse_obj(value)
        return True
    except PydanticValidationError:
        return False


def is_obj(obj: Any) -> bool:
    return obj is not None and isinstance(obj, dict)


def has_prop(data: Dict, prop: str) -> bool:
    return prop in data


class DiscriminatedObject(_BaseLex):
    __root__: Dict[str, Any]

    @validator("__root__")
    def check_type(cls, value):
        if "$type" not in value:
            raise ValueError("DiscriminatedObject must have a $type property")
        return value


def is_discriminated_object(value: Any) -> bool:
    try:
        DiscriminatedObject.parse_obj(value)
        return True
    except ValueError:
        return False


class LexiconDocMalformedError(Exception):
    def __init__(self, message: str, schema_def: Any, issues: Optional[List] = None):
        super().__init__(message)
        self.schema_def = schema_def
        self.issues = issues


class ValidationResult:
    def __init__(
        self, success: bool, value: Optional[Any] = None, error: Optional[Any] = None
    ):
        self.success = success
        self.value = value
        self.error = error


class ValidationError(Exception):
    pass


class InvalidLexiconError(Exception):
    pass


class LexiconDefNotFoundError(Exception):
    pass





import json
from pathlib import Path
from psychonaut.lexicon.types import LexiconDoc


def generate_defs_from(file_path: Path, output_dir: Path, verbose: bool = False):
    # Load the LexiconDoc
    with open(file_path, "r") as json_file:
        doc = LexiconDoc(**json.load(json_file))

    defs = doc.defs
    if not defs:
        return

    print(doc.id)
    
    for k, defn in doc.defs.items():
        if defn.type != "object":
            continue

        props = defn.properties
        if not props:
            continue

        for k, v in props.items():
            if v.type == "ref":
                ref = v.ref
                if ref.startswith("#"):
                    ref = f"{doc.id}{ref}"
                print(f"\tFound ref: {k} -> {ref}")




from typing import Any, Optional
from psychonaut.api.lexicons.com.atproto.server.defs import InviteCode
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetInviteCodesResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    codes: Any


class GetInviteCodesReq(BaseModel):
    """
    Admin view of invite codes
    """

    sort: Optional[str] = Field(default="recent", known_values=["recent", "usage"])
    limit: Optional[int] = Field(default=100, ge=1, le=500)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.getInviteCodes"

    async def do_xrpc(self, sess: Session) -> GetInviteCodesResp:
        resp = await sess.query(self)
        return GetInviteCodesResp(**resp)



from typing import Any, Optional, List
from pydantic import BaseModel, Field
from psychonaut.api.session import Session


class Links(BaseModel):
    """
    [none provided by spec]
    """

    privacyPolicy: Optional[str] = Field(default=None)
    termsOfService: Optional[str] = Field(default=None)


class DescribeServerResp(BaseModel):
    inviteCodeRequired: Optional[bool] = Field(default=None)
    availableUserDomains: List[str] = Field(...)
    links: Optional[Links] = None


class DescribeServerReq(BaseModel):
    """
    Get a document describing the service's accounts configuration.
    """

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.describeServer"

    async def do_xrpc(self, sess: Session) -> DescribeServerResp:
        resp = await sess.query(self)
        return DescribeServerResp(**resp)



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did, validate_datetime


class Block(BaseModel):
    """
    A block.
    """

    subject: str = Field(..., pre=True, validator=validate_did)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.graph.block"



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_datetime
from psychonaut.api.session import Session


class AppPassword(BaseModel):
    """
    [none provided by spec]
    """

    name: str = Field(...)
    createdAt: str = Field(..., pre=True, validator=validate_datetime)


class ListAppPasswordsResp(BaseModel):
    passwords: Any


class ListAppPasswordsReq(BaseModel):
    """
    List all app-specific passwords.
    """

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.server.listAppPasswords"

    async def do_xrpc(self, sess: Session) -> ListAppPasswordsResp:
        resp = await sess.query(self)
        return ListAppPasswordsResp(**resp)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_cid,
    validate_at_identifier,
    validate_nsid,
)


class DeleteRecordReq(BaseModel):
    """
    Delete a record, or ensure it doesn't exist.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    collection: str = Field(
        ...,
        description="The NSID of the record collection.",
        pre=True,
        validator=validate_nsid,
    )
    rkey: str = Field(..., description="The key of the record.")
    swapRecord: Optional[str] = Field(
        default=None,
        description="Compare and swap with the previous record by cid.",
        pre=True,
        validator=validate_cid,
    )
    swapCommit: Optional[str] = Field(
        default=None,
        description="Compare and swap with the previous commit by cid.",
        pre=True,
        validator=validate_cid,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.deleteRecord"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from typing import Any, Optional
from psychonaut.api.lexicons.app.bsky.feed.defs import FeedViewPost
from psychonaut.api.session import Session
from pydantic import BaseModel, Field


class GetTimelineResp(BaseModel):
    cursor: Optional[str] = Field(default=None)
    feed: Any


class GetTimelineReq(BaseModel):
    """
    A view of the user's home timeline.
    """

    algorithm: Optional[str] = Field(default=None)
    limit: Optional[int] = Field(default=50, ge=1, le=100)
    cursor: Optional[str] = Field(default=None)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.getTimeline"

    async def do_xrpc(self, sess: Session) -> GetTimelineResp:
        resp = await sess.query(self)
        return GetTimelineResp(**resp)



from typing import Any, Optional
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import (
    validate_at_uri,
    validate_cid,
    validate_at_identifier,
    validate_nsid,
)


class GetRecordResp(BaseModel):
    uri: str = Field(..., pre=True, validator=validate_at_uri)
    cid: Optional[str] = Field(default=None, pre=True, validator=validate_cid)
    value: Any


class GetRecordReq(BaseModel):
    """
    Get a record.
    """

    repo: str = Field(
        ...,
        description="The handle or DID of the repo.",
        pre=True,
        validator=validate_at_identifier,
    )
    collection: str = Field(
        ...,
        description="The NSID of the record collection.",
        pre=True,
        validator=validate_nsid,
    )
    rkey: str = Field(..., description="The key of the record.")
    cid: Optional[str] = Field(
        default=None,
        description="The CID of the version of the record. If not specified, then return the most recent version.",
        pre=True,
        validator=validate_cid,
    )

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.repo.getRecord"

    async def do_xrpc(self, sess: Session) -> GetRecordResp:
        resp = await sess.query(self)
        return GetRecordResp(**resp)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_did


class GetRepoReq(BaseModel):
    """
    View details about a repository.
    """

    did: str = Field(..., pre=True, validator=validate_did)

    @property
    def xrpc_id(self) -> str:
        return "com.atproto.admin.getRepo"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.query(self)



from typing import Any
from psychonaut.api.session import Session
from pydantic import BaseModel, Field
from psychonaut.lexicon.formats import validate_datetime


class UpdateSeenReq(BaseModel):
    """
    Notify server that the user has seen notifications.
    """

    seenAt: str = Field(..., pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.notification.updateSeen"

    async def do_xrpc(self, sess: Session) -> Any:
        return await sess.procedure(self)



from pathlib import Path

from multiformats import CID
from carbox.message import read_event_pair
from psychonaut.util import load_test_fixture
from base64 import b64decode


# def test_find_b64_line():
#     file_path = Path("/home/generativist/Social/bluesky/psychonaut/streams/stream.1682291107.b64-lines")

#     def predicate(kind, obj):
#         if 'block' not in obj:
#             return False

#         block = obj['block']
#         if not block:
#             return False

#         if 'text' not in block:
#             return False


#         return 'donkeyballs' in block['text']

#     line = find_b64_line(file_path, predicate)
#     print(line)

#     assert line == "d"


def test_read_msg_pair(load_test_fixture):
    post = b64decode(load_test_fixture("donkeyballs.b64"))

    header, event = read_event_pair(post)
    assert header == {"op": 1, "t": "#commit"}

    keys = {
        "ops",
        "seq",
        "prev",
        "repo",
        "time",
        "blobs",
        "blocks",
        "commit",
        "rebase",
        "tooBig",
    }
    assert set(event) == keys

    assert len(event["ops"]) == 1
    op = event["ops"][0]
    assert op["action"] == "create"
    assert op["path"] == "app.bsky.feed.post/3ju35q7husm2p"
    cid = op["cid"]
    assert isinstance(cid, CID)
    assert str(cid) == "zdpuAx7GYAybGShxy9wvkK5eJt6a5G47tz5z5yeFcDqChfYE3"

    assert event["seq"] == 4715462
    assert isinstance(event["prev"], CID)
    assert str(event["prev"]) == "zdpuAmv8cg3iFrz9rn14kYpCZDphscAnYxzxm8EKodWpXbRec"
    assert event["repo"] == "did:plc:o32okshy54r5h2vlrjpz3aln"
    assert event["time"] == "2023-04-23T23:05:15.389Z"

    assert len(event["blobs"]) == 0
    assert len(event["blocks"]) == 5082
    assert isinstance(event["commit"], CID)
    assert str(event["commit"]) == "zdpuArKcqh4Bfc5ufSWKTSS1jFRYJ47gpuxCEVXeWdMEjDpAM"

    assert not event["rebase"]
    assert not event["tooBig"]

    # assert blocks[2] == { }



from typing import Any
from pydantic import BaseModel, Field
from psychonaut.api.lexicons.com.atproto.repo.strong_ref import StrongRef
from psychonaut.lexicon.formats import validate_datetime


class Repost(BaseModel):
    """
    [none provided by spec]
    """

    subject: StrongRef
    createdAt: str = Field(..., pre=True, validator=validate_datetime)

    @property
    def xrpc_id(self) -> str:
        return "app.bsky.feed.repost"



"""
See: https://github.com/bluesky-social/atproto/blob/main/packages/lexicon/src/blob-refs.ts
"""
from __future__ import annotations
from typing import Any, Dict, Optional, Union
from pydantic import BaseModel, Field, validator
from multiformats.cid import CID
from psychonaut.common_web.ipld import ipld_to_json


# TODO: test

class TypedJsonBlobRef(BaseModel):
    type_kludge: str = Field(..., alias='$type', const='blob')
    ref: CID
    mime_type: str = Field(..., alias='mimeType')
    size: int

    class Config:
        extra = "forbid"


class UntypedJsonBlobRef(BaseModel):
    cid: str
    mime_type: str = Field(..., alias='mimeType')

    class Config:
        extra = "forbid"


JsonBlobRef = Union[TypedJsonBlobRef, UntypedJsonBlobRef]


class BlobRef:
    def __init__(self, ref: CID, mime_type: str, size: int, original: JsonBlobRef = None):
        self.ref = ref
        self.mime_type = mime_type
        self.size = size
        self.original = original if original else TypedJsonBlobRef(
            ref=ref, mimeType=mime_type, size=size
        )

    @staticmethod
    def as_blob_ref(obj: Dict[Any, Any]) -> Optional[BlobRef]:
        try:
            typed_json_blob_ref = TypedJsonBlobRef(**obj)
            return BlobRef.from_json_ref(typed_json_blob_ref)
        except ValueError:
            try:
                untyped_json_blob_ref = UntypedJsonBlobRef(**obj)
                return BlobRef.from_json_ref(untyped_json_blob_ref)
            except ValueError:
                return None

    @staticmethod
    def from_json_ref(json: JsonBlobRef) -> BlobRef:
        if isinstance(json, TypedJsonBlobRef):
            return BlobRef(json.ref, json.mimeType, json.size)
        else:
            return BlobRef(CID.from_string(json.cid), json.mimeType, -1, json)

    def ipld(self) -> TypedJsonBlobRef:
        return TypedJsonBlobRef(ref=self.ref, mimeType=self.mime_type, size=self.size)

    def to_json(self) -> dict:
        return ipld_to_json(self.ipld())
